#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought logger package component>
#------------------------------------------------------------------------------

# Standard library imports.
import logging

# Enthought library imports.
from pyface.api import Dialog
from traits.api import Any, Str, Tuple


# Setup a logger for this module.
logger = logging.getLogger(__name__)


priority_levels = ['Low', 'Medium', 'High', 'Critical']


class QualityAgentView(Dialog):

    size = Tuple((700, 900))
    title = Str('Quality Agent')

    # The associated LoggerService.
    service = Any()

    msg = Str('')
    subject = Str('Untitled Error Report')
    to_address = Str()
    cc_address = Str('')
    from_address = Str()
    smtp_server = Str()
    priority = Str(priority_levels[2])
    comments = Str('None')
    include_userdata = Any

    ###########################################################################
    # Protected 'Dialog' interface.
    ###########################################################################

    # fixme: Ideally, this should be passed in; this topic ID belongs to the
    #        Enlib help project/plug-in.
    help_id = 'enlib|HID_Quality_Agent_Dlg'

    def _create_dialog_area(self, parent):
        """ Creates the main content of the dialog. """
        import wx

        parent.SetSizeHints(minW=300, minH=575)

        # Add the main panel
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel = wx.Panel(parent, -1)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)


        # Add a descriptive label at the top ...
        label = wx.StaticText(panel, -1, "Send a comment or bug report ...")
        sizer.Add(label, 0, wx.ALL, border=5)

        # Add the stack trace view ...
        error_panel = self._create_error_panel(panel)
        sizer.Add(error_panel, 1, wx.ALL|wx.EXPAND|wx.CLIP_CHILDREN, border=5)

        # Update the layout:
        sizer.Fit(panel)

        # Add the error report view ...
        report_panel = self._create_report_panel(panel)
        sizer.Add(report_panel, 2, wx.ALL|wx.EXPAND|wx.CLIP_CHILDREN, border=5)

        # Update the layout:
        sizer.Fit(panel)

        return panel


    def _create_buttons(self, parent):
        """ Creates the buttons. """
        import wx

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        # 'Send' button.
        send = wx.Button(parent, wx.ID_OK, "Send")
        wx.EVT_BUTTON(parent, wx.ID_OK, self._on_send)
        sizer.Add(send)
        send.SetDefault()

        # 'Cancel' button.
        cancel = wx.Button(parent, wx.ID_CANCEL, "Cancel")
        wx.EVT_BUTTON(parent, wx.ID_CANCEL, self._wx_on_cancel)
        sizer.Add(cancel, 0, wx.LEFT, 10)

        # 'Help' button.
        if len(self.help_id) > 0:
            help = wx.Button(parent, wx.ID_HELP, "Help")
            wx.EVT_BUTTON(parent, wx.ID_HELP, self._wx_on_help)
            sizer.Add(help, 0, wx.LEFT, 10)

        return sizer

    def _on_help(self, event):
        """Called when the 'Help' button is pressed. """

        hp = self.service.application.get_service('apptools.help.IHelp')
        hp.library.show_topic(self.help_id)

        return


    ### Utility methods #######################################################

    def _create_error_panel(self, parent):
        import wx

        box = wx.StaticBox(parent, -1, "Message:")
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)

        # Print the stack trace
        label2 = wx.StaticText(parent, -1,"The following information will be included in the report:")
        sizer.Add(label2, 0, wx.LEFT|wx.TOP|wx.BOTTOM|wx.CLIP_CHILDREN, border=5)

        details = wx.TextCtrl(parent, -1, self.msg, size=(-1,75),
                              style=wx.TE_MULTILINE |
                                    wx.TE_READONLY |
                                    wx.HSCROLL |
                                    wx.VSCROLL |
                                    wx.TE_RICH2 |
                                    wx.CLIP_CHILDREN)
        details.SetSizeHints(minW=-1, minH=75)
        # Set the font to not be proportional
        font = wx.Font(12, wx.MODERN, wx.NORMAL, wx.NORMAL)
        details.SetStyle(0, len(self.msg), wx.TextAttr(font=font))
        sizer.Add(details, 1, wx.EXPAND|wx.ALL|wx.CLIP_CHILDREN, 5)


        return sizer


    def _create_report_panel(self, parent):
        import wx

        box = wx.StaticBox(parent, -1, "Report Information:")
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)

        # Add email info ...
        sizer.Add(self._create_email_info(parent), 0, wx.ALL|wx.EXPAND, 5)

        # Add priority combo:
        sizer.Add(self._create_priority_combo(parent), 0, wx.ALL|wx.RIGHT, 5)

        # Extra comments from the user:
        label3 = wx.StaticText(parent, -1, "Additional Comments:")
        sizer.Add(label3, 0, wx.LEFT|wx.TOP|wx.BOTTOM|wx.CLIP_CHILDREN, 5)

        comments_field = wx.TextCtrl(parent, -1, self.comments, size=(-1,75),
                                     style=wx.TE_MULTILINE |
                                           wx.TE_RICH2 |
                                           wx.CLIP_CHILDREN)
        comments_field.SetSizeHints(minW=-1, minH=75)
        font = wx.Font(12, wx.MODERN, wx.NORMAL, wx.NORMAL)
        comments_field.SetStyle(0, len(self.comments), wx.TextAttr(font=font))
        sizer.Add(comments_field, 1, wx.ALL|wx.EXPAND|wx.CLIP_CHILDREN, 5)
        wx.EVT_TEXT(parent, comments_field.GetId(), self._on_comments)

        # Include the project combobox?
        if len(self.service.mail_files) > 0:
            sizer.Add(self._create_project_upload(parent), 0, wx.ALL, border=5)

        return sizer


    def _create_email_info(self, parent):
        import wx

        # Layout setup ..
        sizer = wx.FlexGridSizer(5,2,10,10)
        sizer.AddGrowableCol(1)

        title_label = wx.StaticText(parent, -1, "Subject:")
        sizer.Add(title_label , 0, wx.ALL|wx.ALIGN_RIGHT)
        title_field = wx.TextCtrl(parent, -1, self.subject, wx.Point(-1,-1))
        sizer.Add(title_field, 1, wx.EXPAND|wx.ALL|wx.ALIGN_RIGHT|wx.CLIP_CHILDREN)
        wx.EVT_TEXT(parent, title_field.GetId(), self._on_subject)

        to_label = wx.StaticText(parent, -1, "To:")
        sizer.Add(to_label , 0, wx.ALL|wx.ALIGN_RIGHT)
        to_field = wx.TextCtrl(parent, -1, self.to_address)
        sizer.Add(to_field, 1, wx.EXPAND|wx.ALL|wx.ALIGN_RIGHT|wx.CLIP_CHILDREN)
        wx.EVT_TEXT(parent, to_field.GetId(), self._on_to)

        cc_label = wx.StaticText(parent, -1, "Cc:")
        sizer.Add(cc_label, 0, wx.ALL|wx.ALIGN_RIGHT)
        cc_field = wx.TextCtrl(parent, -1, "")
        sizer.Add(cc_field, 1, wx.EXPAND|wx.ALL|wx.ALIGN_RIGHT|wx.CLIP_CHILDREN)
        wx.EVT_TEXT(parent, cc_field.GetId(), self._on_cc)

        from_label = wx.StaticText(parent, -1, "From:")
        sizer.Add(from_label, 0, wx.ALL|wx.ALIGN_RIGHT)
        from_field = wx.TextCtrl(parent, -1, self.from_address)
        sizer.Add(from_field, 1, wx.EXPAND|wx.ALL|wx.ALIGN_RIGHT|wx.CLIP_CHILDREN)
        wx.EVT_TEXT(parent, from_field.GetId(), self._on_from)

        smtp_label = wx.StaticText(parent, -1, "SMTP Server:")
        sizer.Add(smtp_label, 0, wx.ALL|wx.ALIGN_RIGHT)
        smtp_server_field = wx.TextCtrl(parent, -1, self.smtp_server)
        sizer.Add(smtp_server_field, 1, wx.EXPAND|wx.ALL|wx.ALIGN_RIGHT|wx.CLIP_CHILDREN)
        wx.EVT_TEXT(parent, smtp_server_field.GetId(), self._on_smtp_server)

        return sizer


    def _create_priority_combo(self, parent):
        import wx

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        label = wx.StaticText(parent, -1, "How critical is this issue?")
        sizer.Add(label, 0, wx.ALL, border=0)

        cb = wx.ComboBox(parent, -1, self.priority,
                         wx.Point(90, 50), wx.Size(95, -1),
                         priority_levels, wx.CB_READONLY)
        sizer.Add(cb, 1, wx.EXPAND|wx.LEFT|wx.CLIP_CHILDREN, border=10)

        wx.EVT_COMBOBOX(parent, cb.GetId(), self._on_priority)

        return sizer


    def _create_project_upload(self, parent):
        import wx

        id = wx.NewId()
        cb = wx.CheckBox(parent, id, "Include Workspace Files (will increase email size)        ",
                                wx.Point(65, 80), wx.Size(-1, 20), wx.NO_BORDER)
        wx.EVT_CHECKBOX(parent, id, self._on_project)

        return cb


    ## UI Listeners ###########################################################

    def _on_subject(self, event):
        self.subject = event.GetEventObject().GetValue()


    def _on_to(self, event):
        self.to_address = event.GetEventObject().GetValue()


    def _on_cc(self, event):
        self.cc_address = event.GetEventObject().GetValue()


    def _on_from(self, event):
        self.from_address = event.GetEventObject().GetValue()


    def _on_smtp_server(self, event):
        self.smtp_server = event.GetEventObject().GetValue()


    def _on_priority(self, event):
        self.priority = event.GetEventObject().GetStringSelection()


    def _on_comments(self, event):
        self.comments = event.GetEventObject().GetValue()


    def _on_project(self, event):
        self.include_userdata = event.Checked()
        cb = event.GetEventObject()

        if event.Checked():
            cb.SetLabel("Include Workspace Files (approx. %.2f MBytes)" % self._compute_project_size())
        else:
            cb.SetLabel("Include Workspace Files (will increase email size)")
        return

    def _on_send(self, event):
        import wx
        # Disable the Send button while we go through the possibly
        # time-consuming email-sending process.
        button = event.GetEventObject()
        button.Enable(0)

        fromaddr, toaddrs, ccaddrs = self._create_email_addresses()
        message = self._create_email(fromaddr, toaddrs, ccaddrs)

        self.service.send_bug_report(self.smtp_server, fromaddr, toaddrs,
            ccaddrs, message)

        # save the user's preferences
        self.service.preferences.smtp_server = self.smtp_server
        self.service.preferences.to_address = self.to_address
        self.service.preferences.from_address = self.from_address

        # finally we close the dialog
        self._wx_on_ok(event)

        return

    ## Private ################################################################

    def _create_email_addresses(self):
        # utility function map addresses from ui into the standard format
        # FIXME: We should use standard To: header parsing instead of this ad
        # hoc whitespace-only approach.
        fromaddr = self.from_address
        if "" == fromaddr.strip():
            fromaddr = "anonymous"
        toaddrs = self.to_address.split()
        ccaddrs = self.cc_address.split()

        return fromaddr, toaddrs, ccaddrs


    def _compute_project_size(self):
        # determine size of email in MBytes
        fromaddr, toaddrs, ccaddrs = self._create_email_addresses()
        message = self._create_email(fromaddr, toaddrs, ccaddrs)
        return len(message.as_string()) / (2.0**20)


    def _create_email(self, fromaddr, toaddrs, ccaddrs):
        return self.service.create_email_message(
            fromaddr, toaddrs, ccaddrs,
            self.subject,
            self.priority,
            self.include_userdata,
            self.msg,
            self.comments,
        )

    def _to_address_default(self):
        return self.service.preferences.to_address

    def _from_address_default(self):
        return self.service.preferences.from_address

    def _smtp_server_default(self):
        return self.service.preferences.smtp_server

####### EOF #############################################################



from traits.api import Any, Event, HasTraits, List, provides, Str
from traits.testing.unittest_tools import unittest

from apptools.selection.api import (
    IDConflictError, ISelection, ISelectionProvider, ListenerNotConnectedError,
    ListSelection, ProviderNotRegisteredError, SelectionService)


@provides(ISelection)
class BogusSelection(HasTraits):

    provider_id = Str

    # Some content to check that two selections are the same
    content = Any

    def is_empty(self):
        """ Is the selection empty? """
        return False


@provides(ISelectionProvider)
class BogusSelectionProvider(HasTraits):

    #### 'ISelectionProvider' protocol ########################################

    provider_id = Str

    selection = Event

    def get_selection(self):
        return BogusSelection(provider_id=self.provider_id,
                              content='get_selection')

    def set_selection(self, items, ignore_missing=False):
        pass

    #### 'BogusSelectionProvider' protocol ####################################

    def trigger_selection(self, content):
        self.selection = BogusSelection(provider_id=self.provider_id,
                                        content=content)


class BogusListener(HasTraits):
    selections = List

    def on_selection_changed(self, selection):
        self.selections.append(selection)


@provides(ISelectionProvider)
class SimpleListProvider(HasTraits):

    #### 'ISelectionProvider' protocol ########################################

    provider_id = Str('test.simple_list_provider')

    selection = Event

    def get_selection(self):
        selection = ListSelection.from_available_items(
            provider_id=self.provider_id,
            selected=self._selected,
            all_items=self.items
        )
        return selection

    def set_selection(self, items, ignore_missing=False):
        selected = [x for x in items if x in self.items]
        if not ignore_missing and len(selected) < len(items):
            raise ValueError()
        self._selected = selected

    #### 'SimpleListProvider' protocol ########################################

    items = List

    _selected = List


class TestSelectionService(unittest.TestCase):

    def test_add_selection_provider(self):
        service = SelectionService()
        provider = BogusSelectionProvider()

        service.add_selection_provider(provider)
        self.assertTrue(service.has_selection_provider(provider.provider_id))

    def test_add_selection_id_conflict(self):
        service = SelectionService()

        provider_id = 'Foo'
        provider = BogusSelectionProvider(provider_id=provider_id)
        another_provider = BogusSelectionProvider(provider_id=provider_id)

        service.add_selection_provider(provider)
        with self.assertRaises(IDConflictError):
            service.add_selection_provider(another_provider)

    def test_remove_selection_provider(self):
        service = SelectionService()
        provider = BogusSelectionProvider(provider_id='Bogus')

        service.add_selection_provider(provider)
        service.remove_selection_provider(provider)
        self.assertFalse(service.has_selection_provider(provider.provider_id))

        with self.assertRaises(ProviderNotRegisteredError):
            service.remove_selection_provider(provider)

    def test_get_selection(self):
        service = SelectionService()
        provider_id = 'Bogus'
        provider = BogusSelectionProvider(provider_id=provider_id)
        service.add_selection_provider(provider)

        selection = service.get_selection(provider_id)

        self.assertIsInstance(selection, ISelection)
        self.assertEqual(selection.provider_id, provider.provider_id)

    def test_get_selection_id_not_registered(self):
        service = SelectionService()

        with self.assertRaises(ProviderNotRegisteredError):
            service.get_selection('not-registered')

    def test_connect_listener(self):
        service = SelectionService()
        provider_id = 'Bogus'
        provider = BogusSelectionProvider(provider_id=provider_id)
        service.add_selection_provider(provider)

        listener = BogusListener()
        service.connect_selection_listener(provider_id,
                                           listener.on_selection_changed)

        content = [1, 2, 3]
        provider.trigger_selection(content)

        selections = listener.selections
        self.assertEqual(len(selections), 1)
        self.assertEqual(selections[0].provider_id, provider.provider_id)
        self.assertEqual(selections[0].content, content)

    def test_connect_listener_then_add_remove_provider(self):
        service = SelectionService()
        provider_id = 'Bogus'

        # Connect listener before provider is registered.
        listener = BogusListener()
        service.connect_selection_listener(provider_id,
                                           listener.on_selection_changed)

        # When the provider is first added, the listener should receive the
        # initial selection (as returned by provider.get_selection)
        provider = BogusSelectionProvider(provider_id=provider_id)
        expected = provider.get_selection()
        service.add_selection_provider(provider)

        selections = listener.selections
        self.assertEqual(len(selections), 1)
        self.assertEqual(selections[-1].content, expected.content)

        # When the provider changes the selection, the event arrive as usual.
        content = [1, 2, 3]
        provider.trigger_selection(content)

        self.assertEqual(len(selections), 2)
        self.assertEqual(selections[-1].content, content)

        # When we un-register the provider, a change in selection does not
        # generate a callback.
        service.remove_selection_provider(provider)
        provider.trigger_selection(content)
        self.assertEqual(len(selections), 2)

        # Finally, we register again and get the current selection.
        service.add_selection_provider(provider)
        self.assertEqual(len(selections), 3)
        self.assertEqual(selections[-1].content, expected.content)

    def test_disconnect_listener(self):
        service = SelectionService()
        provider_id = 'Bogus'
        provider = BogusSelectionProvider(provider_id=provider_id)
        service.add_selection_provider(provider)

        listener = BogusListener()
        service.connect_selection_listener(provider_id,
                                           listener.on_selection_changed)
        service.disconnect_selection_listener(provider_id,
                                              listener.on_selection_changed)

        provider.trigger_selection([1, 2, 3])

        self.assertEqual(len(listener.selections), 0)

    def test_disconnect_unknown_listener(self):
        service = SelectionService()
        provider_id = 'Bogus'
        provider = BogusSelectionProvider(provider_id=provider_id)
        service.add_selection_provider(provider)

        # First case: there are listeners to a provider, but not the one we
        # pass to the disconnect method
        listener_1 = BogusListener()
        service.connect_selection_listener(provider_id,
                                           listener_1.on_selection_changed)

        listener_2 = BogusListener()
        with self.assertRaises(ListenerNotConnectedError):
            service.disconnect_selection_listener(
                provider_id, listener_2.on_selection_changed)

        # Second case: there is no listener connected to the ID
        with self.assertRaises(ListenerNotConnectedError):
            service.disconnect_selection_listener(
                'does-not-exists', listener_2.on_selection_changed)

    def test_set_selection(self):
        service = SelectionService()
        provider = SimpleListProvider(items=list(range(10)))
        service.add_selection_provider(provider)

        provider_id = provider.provider_id
        selection = service.get_selection(provider_id)
        self.assertTrue(selection.is_empty())

        new_selection = [5, 6, 3]
        service.set_selection(provider_id, new_selection)

        selection = service.get_selection(provider_id)
        self.assertFalse(selection.is_empty())

        # We can't assume that the order of the items in the selection we set
        # remains stable.
        self.assertEqual(selection.items, new_selection)
        self.assertEqual(selection.indices, selection.items)

    def test_selection_id_not_registered(self):
        service = SelectionService()

        with self.assertRaises(ProviderNotRegisteredError):
            service.set_selection(provider_id='not-existent', items=[])

    def test_ignore_missing(self):
        # What we are really testing here is that the selection service
        # passes the keyword argument to the selection provider. It's the
        # selection provider responsibility to ignore missing elements, or
        # raise an exception.

        service = SelectionService()
        provider = SimpleListProvider(items=list(range(10)))
        service.add_selection_provider(provider)

        new_selection = [0, 11, 1]
        provider_id = provider.provider_id
        service.set_selection(provider_id, new_selection, ignore_missing=True)

        selection = service.get_selection(provider_id)
        self.assertFalse(selection.is_empty())
        self.assertEqual(selection.items, [0, 1])

        new_selection = [0, 11, 1]
        with self.assertRaises(ValueError):
            service.set_selection(provider_id, new_selection,
                                  ignore_missing=False)


if __name__ == '__main__':
    unittest.main()



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought application scripting package component>
#------------------------------------------------------------------------------


# Standard library imports.
import datetime
import types
import weakref

# Enthought library imports.
from traits.api import Any, Bool, Callable, Dict, Event, HasTraits, \
        implements, Instance, Int, List, Property, Str, Unicode

# Local imports.
from .bind_event import BindEvent
from .i_bind_event import IBindEvent
from .i_script_manager import IScriptManager
from .lazy_namespace import add_to_namespace, FactoryWrapper, LazyNamespace
from .scriptable_type import make_object_scriptable


@provides(IScriptManager)
class _BoundObject(HasTraits):
    """The base class for any object that can be bound to a name."""

    #### '_BoundObject' interface #############################################

    # Set if the object was explicitly bound.
    explicitly_bound = Bool(True)

    # The name the object is bound to.
    name = Str

    # The object being bound.
    obj = Any


class _ScriptObject(_BoundObject):
    """The _ScriptObject class encapsulates a scriptable object."""

    #### '_BoundObject' interface #############################################

    # The object being bound.
    obj = Property

    #### '_ScriptObject' interface ############################################

    # The positional arguments passed to __init__ after being converted to
    # strings.  A particular argument may be an exception if it couldn't be
    # converted.
    args = List

    # The keyword arguments passed to __init__ after being converted to
    # strings.  A particular argument may be an exception if it couldn't be
    # converted.
    kwargs = Dict

    # The id of the object.
    obj_id = Int

    # A weak reference to the object.
    obj_ref = Any

    # The type of the scriptable object.
    scripted_type = Any

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_obj(self):
        """The property getter."""

        return self.obj_ref()


class _ScriptCall(HasTraits):
    """ The _ScriptCall class is the base class for all script calls. """

    #### '_ScriptCall' interface ##############################################

    # The name of the call.
    name = Str

    # The scriptable object.
    so = Any

    ###########################################################################
    # '_ScriptCall' interface.
    ###########################################################################

    def as_str(self, sm, so_needed):
        """ Return the string equivalent of the call, updated the list of
        needed scriptable objects if required.
        """

        raise NotImplementedError


class _ScriptTraitGet(_ScriptCall):
    """ The _ScriptTraitGet class encapsulates a single call to the get of a
    scriptable trait.
    """

    #### '_ScriptTraitGet' interface ##########################################

    # Set if the getter has side effects.
    has_side_effects = Bool(False)

    # The result of the get.  Keeping a reference to it means that the memory
    # can't get reused.
    result = Any

    ###########################################################################
    # '_ScriptCall' interface.
    ###########################################################################

    def as_str(self, sm, so_needed):
        """ Return the string equivalent of the call, updated the list of
        needed scriptable objects if required.
        """

        # Ignore if it is no longer bound.
        if not self.so.name:
            return None

        if self.result is None:
            rstr = ""
        else:
            nr, _ = sm._results[id(self.result)]

            if nr >= 0:
                rstr = "r%d = " % nr
            else:
                rstr = ""

        # Unless getter has side effects, if the result is not needed then
        # don't bother including it in the script.
        if not self.has_side_effects and rstr == "":
            return None

        so = sm.arg_as_string(self.so, so_needed)

        return "%s%s.%s" % (rstr, so, self.name)


class _ScriptTraitSet(_ScriptCall):
    """ The _ScriptTraitSet class encapsulates a single call to the set of a
    scriptable trait.
    """

    #### '_ScriptTraitSet' interface ##########################################

    # The value the trait is set to.
    value = Any

    ###########################################################################
    # '_ScriptCall' interface.
    ###########################################################################

    def as_str(self, sm, so_needed):
        """ Return the string equivalent of the call, updated the list of
        needed scriptable objects if required.
        """

        # Ignore if it is no longer bound.
        if not self.so.name:
            return None

        so = sm.arg_as_string(self.so, so_needed)
        value = sm.arg_as_string(self.value, so_needed)

        return "%s.%s = %s" % (so, self.name, value)


class _ScriptMethod(_ScriptCall):
    """ The _ScriptMethod class encapsulates a single call to a scriptable
    method.
    """

    #### '_ScriptMethod' interface ############################################

    # The positional arguments passed to the method after being converted to
    # strings.  A particular argument may be an exception if it couldn't be
    # converted.
    args = List

    # The keyword arguments passed to the method after being converted to
    # strings.  A particular argument may be an exception if it couldn't be
    # converted.
    kwargs = Dict

    # The result of the method call.  Keeping a reference to it means that the
    # memory can't get reused.
    result = Any

    ###########################################################################
    # '_ScriptCall' interface.
    ###########################################################################

    def as_str(self, sm, so_needed):
        """ Return the string equivalent of the call, updated the list of
        needed scriptable objects if required.
        """

        # Ignore if it is no longer bound.
        if self.so and not self.so.name:
            return None

        if self.result is None:
            rstr = ""
        elif type(self.result) is type(()):
            rlist = []
            needed = False

            for r in self.result:
                nr, _ = sm._results[id(r)]

                if nr >= 0:
                    rlist.append("r%d" % nr)
                    needed = True
                else:
                    rlist.append("_")

            if needed:
                rstr = ", ".join(rlist) + " = "
            else:
                rstr = ""
        else:
            nr, _ = sm._results[id(self.result)]

            if nr >= 0:
                rstr = "r%d = " % nr
            else:
                rstr = ""

        if self.so:
            so = sm.arg_as_string(self.so, so_needed) + '.'
        else:
            so = ''

        args = sm.args_as_string_list(self.args, self.kwargs, so_needed)

        return "%s%s%s(%s)" % (rstr, so, self.name, ", ".join(args))


class _FactoryObject(_BoundObject):
    """ The _FactoryObject class wraps a factory that lazily creates
    scriptable objects.
    """

    #### '_BoundObject' interface #############################################

    # The object being bound.
    obj = Property

    #### '_FactoryObject' interface ###########################################

    # The optional object that defines the scripting API.
    api = Any

    # The scriptable object factory.
    factory = Callable

    # The optional attribute include list.
    includes = Any

    # The optional attribute exclude list.
    excludes = Any

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_obj(self):
        """The property getter."""

        return FactoryWrapper(factory=self.factory, api=self.api,
                includes=self.includes, excludes=self.excludes)


class ScriptManager(HasTraits):
    """ The ScriptManager class is the default implementation of
    IScriptManager.
    """



    #### 'IScriptManager' interface ###########################################

    # This event is fired whenever a scriptable object is bound or unbound.  It
    # is intended to be used by an interactive Python shell to give the
    # advanced user access to the scriptable objects.  If an object is created
    # via a factory then the event is fired when the factory is called, and not
    # when the factory is bound.
    bind_event = Event(IBindEvent)

    # This is set if user actions are being recorded as a script.  It is
    # maintained by the script manager.
    recording = Bool(False)

    # This is the text of the script currently being recorded (or the last
    # recorded script if none is currently being recorded).  It is updated
    # automatically as the user performs actions.
    script = Property(Unicode)

    # This event is fired when the recorded script changes.  The value of the
    # event will be the ScriptManager instance.
    script_updated = Event(IScriptManager)

    #### Private interface ####################################################

    # The list of calls to scriptable calls.
    _calls = List(Instance(_ScriptCall))

    # The dictionary of bound names.  The value is the next numerical suffix
    # to use when the binding policy is 'auto'.
    _names = Dict

    # The dictionary of _BoundObject instances keyed by the name the object is
    # bound to.
    _namespace = Dict

    # The next sequential result number.
    _next_result_nr = Int

    # The results returned by previous scriptable calls.  The key is the id()
    # of the result object.  The value is a two element tuple of the sequential
    # result number (easier for the user to use than the id()) and the result
    # object itself.
    _results = Dict

    # The dictionary of _ScriptObject instances keyed by the object's id().
    _so_by_id = Dict

    # The dictionary of _ScriptObject instances keyed by the a weak reference
    # to the object.
    _so_by_ref = Dict

    # The date and time when the script was recorded.
    _when_started = Any

    ###########################################################################
    # 'IScriptManager' interface.
    ###########################################################################

    def bind(self, obj, name=None, bind_policy='unique', api=None,
            includes=None, excludes=None):
        """ Bind obj to name and make (by default) its public methods and
        traits (ie. those not beginning with an underscore) scriptable.  The
        default value of name is the type of obj with the first character
        forced to lower case.  name may be a dotted name (eg. 'abc.def.xyz').

        bind_policy determines what happens if the name is already bound.  If
        the policy is 'auto' then a numerical suffix will be added to the name,
        if necessary, to make it unique.  If the policy is 'unique' then an
        exception is raised.  If the policy is 'rebind' then the previous
        binding is discarded.  The default is 'unique'

        If api is given then it is a class, or a list of classes, that define
        the attributes that will be made scriptable.

        Otherwise if includes is given it is a list of names of attributes that
        will be made scriptable.

        Otherwise all the public attributes of scripted_type will be made
        scriptable except those in the excludes list.
        """

        # Register the object.
        self.new_object(obj, obj.__class__, name=name, bind_policy=bind_policy)

        # Make it scriptable.
        make_object_scriptable(obj, api=api, includes=includes,
                excludes=excludes)

    def bind_factory(self, factory, name, bind_policy='unique', api=None,
            includes=None, excludes=None):
        """ Bind factory to name.  This does the same as the bind() method
        except that it uses a factory that will be called later on to create
        the object only if the object is needed.

        See the documentation for bind() for a description of the remaining
        arguments.
        """

        name = self._unique_name(name, bind_policy)
        self._namespace[name] = _FactoryObject(name=name, factory=factory,
                api=api, includes=includes, excludes=excludes)

    def run(self, script):
        """ Run the given script, either a string or a file-like object.
        """

        # Initialise the namespace with all explicitly bound objects.
        nspace = LazyNamespace()
        for name, bo in self._namespace.items():
            if bo.explicitly_bound:
                add_to_namespace(bo.obj, name, nspace)

        exec(script, nspace)

    def run_file(self, file_name):
        """ Run the given script file.
        """

        f = open(file_name)
        self.run(f)
        f.close()

    def start_recording(self):
        """ Start the recording of user actions.  The 'script' trait is cleared
        and all subsequent actions are added to 'script'.  The 'recording'
        trait is updated appropriately.
        """

        self._calls = []
        self._next_result_nr = 0
        self._results = {}

        self.recording = True
        self.script_updated = self

    def stop_recording(self):
        """ Stop the recording of user actions.  The 'recording' trait is
        updated appropriately.
        """

        self.recording = False

    ###########################################################################
    # 'ScriptManager' interface.
    ###########################################################################

    def record_method(self, func, args, kwargs):
        """ Record the call of a method of a ScriptableObject instance and
        return the result.  This is intended to be used only by the scriptable
        decorator.
        """
        if self.recording:
            # Record the arguments before the function has a chance to modify
            # them.
            srec = self._new_method(func, args, kwargs)
            result = func(*args, **kwargs)
            self._add_method(srec, result)

            self.script_updated = self
        else:
            result = func(*args, **kwargs)

        return result

    def record_trait_get(self, obj, name, result):
        """ Record the get of a trait of a scriptable object.  This is intended
        to be used only by the Scriptable trait getter.
        """

        if self.recording:
            side_effects = self._add_trait_get(obj, name, result)

            # Don't needlessly fire the event if there are no side effects.
            if side_effects:
                self.script_updated = self

    def record_trait_set(self, obj, name, value):
        """ Record the set of a trait of a scriptable object.  This is intended
        to be used only by the Scriptable trait getter.
        """

        if self.recording:
            self._add_trait_set(obj, name, value)

            self.script_updated = self

    def new_object(self, obj, scripted_type, args=None, kwargs=None, name=None,
            bind_policy='auto'):
        """ Register a scriptable object and the arguments used to create it.
        If no arguments were provided then assume the object is being
        explicitly bound.
        """

        # The name defaults to the type name.
        if not name:
            name = scripted_type.__name__
            name = name[0].lower() + name[1:]

        name = self._unique_name(name, bind_policy)

        obj_id = id(obj)
        obj_ref = weakref.ref(obj, self._gc_script_obj)

        so = _ScriptObject(name=name, obj_id=obj_id, obj_ref=obj_ref,
                scripted_type=scripted_type)

        # If we are told how to create the object then it must be implicitly
        # bound.
        if args is not None:
            # Convert each argument to its string representation if possible.
            # Doing this now avoids problems with mutable arguments.
            so.args = [self._scriptable_object_as_string(a) for a in args]

            for n, value in kwargs.items():
                so.kwargs[n] = self._scriptable_object_as_string(value)

            so.explicitly_bound = False

        # Remember the scriptable object via the different access methods.
        self._so_by_id[obj_id] = so
        self._so_by_ref[obj_ref] = so
        self._namespace[name] = so

        # Note that if anything listening to this event doesn't use weak
        # references then the object will be kept alive.
        self.bind_event = BindEvent(name=name, obj=obj)

    @staticmethod
    def args_as_string_list(args, kwargs, so_needed=None):
        """ Return a complete argument list from sets of positional and keyword
        arguments.  Update the optional so_needed list for those arguments that
        refer to a scriptable object.
        """

        if so_needed is None:
            so_needed = []

        all_args = []

        for arg in args:
            s = ScriptManager.arg_as_string(arg, so_needed)
            all_args.append(s)

        for name, value in kwargs.items():
            s = ScriptManager.arg_as_string(value, so_needed)
            all_args.append('%s=%s' % (name, s))

        return all_args

    @staticmethod
    def arg_as_string(arg, so_needed):
        """ Return the string representation of an argument.  Update the
        so_needed list if the argument refers to a scriptable object.  Any
        delayed conversion exception is handled here.
        """

        if isinstance(arg, Exception):
            raise arg

        if isinstance(arg, _ScriptObject):
            # Check it hasn't been unbound.
            if not arg.name:
                raise NameError("%s has been unbound but is needed by the script" % arg.obj_ref())

            # Add it to the needed list if it isn't already there.
            if arg not in so_needed:
                so_needed.append(arg)

            arg = arg.name

        return arg

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _new_method(self, func, args, kwargs):
        """ Return an object that encapsulates a call to a scriptable method.
        _add_method() must be called to add it to the current script.
        """

        # Convert each argument to its string representation if possible.
        # Doing this now avoids problems with mutable arguments.
        nargs = [self._object_as_string(arg) for arg in args]

        if type(func) is types.FunctionType:
            so = None
        else:
            so = nargs[0]
            nargs = nargs[1:]

        nkwargs = {}
        for name, value in kwargs.items():
            nkwargs[name] = self._object_as_string(value)

        return _ScriptMethod(name=func.__name__, so=so, args=nargs,
                kwargs=nkwargs)

    def _add_method(self, entry, result):
        """ Add a method call (returned by _new_method()), with it's associated
        result and ID, to the current script.
        """

        self._start_script()

        if result is not None:
            # Assume that a tuple represents multiple returned values - not
            # necessarily a valid assumption unless we make it a rule for
            # scriptable functions.
            if type(result) is type(()):
                for r in result:
                    self._save_result(r)
            else:
                self._save_result(result)

            entry.result = result

        self._calls.append(entry)

    def _add_trait_get(self, obj, name, result):
        """ Add a call to a trait getter, with it's associated result and ID,
        to the current script.  Return True if the get had side effects.
        """

        self._start_script()

        side_effects = obj.trait(name).has_side_effects

        if side_effects is None:
            side_effects = False

        so = self._object_as_string(obj)

        if result is not None:
            self._save_result(result)

        self._calls.append(_ScriptTraitGet(so=so, name=name, result=result,
                has_side_effects=side_effects))

        return side_effects

    def _add_trait_set(self, obj, name, value):
        """ Add a call to a trait setter, with it's associated value and ID,
        to the current script.
        """

        self._start_script()

        so = self._object_as_string(obj)
        value = self._object_as_string(value)

        self._calls.append(_ScriptTraitSet(so=so, name=name, value=value))

    def _unique_name(self, name, bind_policy):
        """ Return a name that is guaranteed to be unique according to the bind
        policy.
        """

        # See if the name is already is use.
        bo = self._namespace.get(name)

        if bo is None:
            self._names[name] = 1
        elif bind_policy == 'auto':
            suff = self._names[name]
            self._names[name] = suff + 1

            name = '%s%d' % (name, suff)
        elif bind_policy == 'rebind':
            self._unbind(bo)
        else:
            raise NameError("\"%s\" is already bound to a scriptable object" % name)

        return name

    def _unbind(self, bo):
        """Unbind the given bound object."""

        # Tell everybody it is no longer bound.  Don't bother if it is a
        # factory because the corresponding bound event wouldn't have been
        # fired.
        if not isinstance(bo, _FactoryObject):
            self.bind_event = BindEvent(name=bo.name, obj=None)

        # Forget about it.
        del self._namespace[bo.name]
        bo.name = ''

    @staticmethod
    def _gc_script_obj(obj_ref):
        """ The callback invoked when a scriptable object is garbage collected.
        """

        # Avoid recursive imports.
        from .package_globals import get_script_manager

        sm = get_script_manager()
        so = sm._so_by_ref[obj_ref]

        if so.name:
            sm._unbind(so)

        del sm._so_by_id[so.obj_id]
        del sm._so_by_ref[so.obj_ref]

    def _start_script(self):
        """ Save when a script recording is started. """

        if len(self._calls) == 0:
            self._when_started = datetime.datetime.now().strftime('%c')

    def _object_as_string(self, obj):
        """ Convert an object to a string as it will appear in a script.  An
        exception may be returned (not raised) if there was an error in the
        conversion.
        """

        obj_id = id(obj)

        # See if the argument is the result of a previous call.
        nr, _ = self._results.get(obj_id, (None, None))

        if nr is not None:
            if nr < 0:
                nr = self._next_result_nr
                self._next_result_nr += 1

                # Key on the ID of the argument (which is hashable) rather than
                # the argument itself (which might not be).
                self._results[obj_id] = (nr, obj)

            return "r%d" % nr

        return self._scriptable_object_as_string(obj)

    def _scriptable_object_as_string(self, obj):
        """ Convert an object to a string as it will appear in a script.  An
        exception may be returned (not raised) if there was an error in the
        conversion.
        """

        obj_id = id(obj)

        # If it is a scriptable object we return the object and convert it to a
        # string later when we know it is really needed.
        so = self._so_by_id.get(obj_id)

        if so is not None:
            return so

        # Use the repr result if it doesn't appear to be the generic response,
        # ie. it doesn't contain its own address as a hex string.
        s = repr(obj)

        if hex(obj_id) not in s:
            return s

        # We don't know how to represent the argument as a string.  This is
        # most likely because an appropriate __init__ hasn't been made
        # scriptable.  We don't raise an exception until the user decides to
        # convert the calls to a script.
        return ValueError("unable to create a script representation of %s" % obj)

    def _save_result(self, result):
        """ Save the result of a call to a scriptable method so that it can be
        recognised later.
        """

        if id(result) not in self._results:
            self._results[id(result)] = (-1, result)

    def _get_script(self):
        """ Convert the current list of calls to a script. """

        # Handle the trivial case.
        if len(self._calls) == 0:
            return ""

        # Generate the header.
        header = "# Script generated %s" % self._when_started

        # Generate the calls.
        so_needed = []
        calls = []

        for call in self._calls:
            s = call.as_str(self, so_needed)

            if s:
                calls.append(s)

        calls = "\n".join(calls)

        # Generate the scriptable object constructors.
        types_needed = []
        ctors = []

        for so in so_needed:
            if so.explicitly_bound:
                continue

            so_type = so.scripted_type
            args = self.args_as_string_list(so.args, so.kwargs)

            ctors.append("%s = %s(%s)" % (so.name, so_type.__name__, ", ".join(args)))

            # See if a new import is needed.
            if so_type not in types_needed:
                types_needed.append(so_type)

        ctors = "\n".join(ctors)

        # Generate the import statements.
        imports = []

        for so_type in types_needed:
            imports.append("from %s import %s" % (so_type.__module__, so_type.__name__))

        imports = "\n".join(imports)

        return "\n\n".join([header, imports, ctors, calls]) + "\n"



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" A naming system explorer. """


# Enthought library imports.
from apptools.naming.api import Binding, PyContext
from pyface.api import PythonShell, SplitApplicationWindow
from traits.api import Float, Instance, Str

# Local imports.
from .naming_tree import NamingTree


# Factory function for exploring a Python namespace.
def explore(obj):
    """ View a Python object as a naming context. """

    root = Binding(name='root', obj=PyContext(namespace=obj))

    explorer = Explorer(root=root, size=(1200, 400))
    explorer.open()

    return


class Explorer(SplitApplicationWindow):
    """ The main application window. """

    #### 'Window' interface ###################################################

    title = Str('Naming System Explorer')

    #### 'SplitApplicationWindow' interface ###################################

    # The direction in which the panel is split.
    direction = Str('vertical')

    # The ratio of the size of the left/top pane to the right/bottom pane.
    ratio = Float(0.3)

    # The root binding (usually a binding to a context!).
    root = Instance(Binding)

    ###########################################################################
    # Protected 'SplitApplicationWindow' interface.
    ###########################################################################

    def _create_lhs(self, parent):
        """ Creates the left hand side or top depending on the style. """

        return self._create_tree(parent, self.root)

    def _create_rhs(self, parent):
        """ Creates the panel containing the selected preference page. """

        return self._create_python_shell(parent)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_tree(self, parent, root):
        """ Creates the tree. """

        self._tree = tree = NamingTree(parent, root=root)

        return tree.control

    def _create_python_shell(self, parent):
        """ Creates the Python shell. """

        self._python_shell = python_shell = PythonShell(parent)

        # Bind useful names.
        python_shell.bind('widget', self._tree)
        python_shell.bind('w', self._tree)
        python_shell.bind('window', self)
        python_shell.bind('explore', explore)

        # Execute useful commands to bind useful names ;^)
        python_shell.execute_command('from apptools.naming.api import *')

        return python_shell.control

##### EOF #####################################################################



"""Tests for the file_path module.

"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2005, Enthought, Inc.
# License: BSD Style.

# Standard library imports.
import unittest
import os
import sys
from os.path import abspath, dirname, basename, join
from io import BytesIO

# Enthought library imports.
from apptools.persistence import state_pickler
from apptools.persistence import file_path

class Test:
    def __init__(self):
        self.f = file_path.FilePath()


class TestFilePath(unittest.TestCase):
    def test_relative(self):
        """Test if relative paths are set correctly.
        """
        fname = 't.vtk'
        f = file_path.FilePath(fname)
        cwd = os.getcwd()
        # Trivial case of both in same dir.
        f.set_relative(abspath(join(cwd, 't.mv2')))
        self.assertEqual(f.rel_pth, fname)
        # Move one directory deeper.
        f.set_relative(abspath(join(cwd, 'tests', 't.mv2')))
        self.assertEqual(f.rel_pth, join(os.pardir, fname))
        # Move one directory shallower.
        f.set_relative(abspath(join(dirname(cwd), 't.mv2')))
        diff = basename(cwd)
        self.assertEqual(f.rel_pth, join(diff, fname))
        # Test where the path is relative to the root.
        f.set(abspath(join('data', fname)))
        f.set_relative('/tmp/test.mv2')
        if sys.platform.startswith('win'):
            expect = os.pardir + abspath(join('data', fname))[2:]
        else:
            expect = os.pardir + abspath(join('data', fname))
        self.assertEqual(f.rel_pth, expect)

    def test_absolute(self):
        """Test if absolute paths are set corectly.
        """
        fname = 't.vtk'
        f = file_path.FilePath(fname)
        cwd = os.getcwd()
        # Easy case of both in same dir.
        f.set_absolute(join(cwd, 'foo', 'test', 't.mv2'))
        self.assertEqual(f.abs_pth, join(cwd, 'foo', 'test', fname))
        # One level lower.
        fname = join(os.pardir, 't.vtk')
        f.set(fname)
        f.set_absolute(join(cwd, 'foo', 'test', 't.mv2'))
        self.assertEqual(f.abs_pth, abspath(join(cwd, 'foo', 'test', fname)))
        # One level higher.
        fname = join('test', 't.vtk')
        f.set(fname)
        f.set_absolute(join(cwd, 'foo', 't.mv2'))
        self.assertEqual(f.abs_pth, abspath(join(cwd, 'foo', fname)))

    def test_pickle(self):
        """Test if pickler works correctly with FilePaths.
        """
        t = Test()
        t.f.set('t.vtk')
        cwd = os.getcwd()
        curdir = basename(cwd)

        # Create a dummy file in the parent dir.
        s = BytesIO()
        # Spoof its location.
        s.name = abspath(join(cwd, os.pardir, 't.mv2'))
        # Dump into it
        state_pickler.dump(t, s)

        # Rewind the stream
        s.seek(0)
        # "Move" the file elsewhere
        s.name = join(cwd, 'foo', 'test', 't.mv2')
        state = state_pickler.load_state(s)
        self.assertEqual(state.f.abs_pth,
                         join(cwd, 'foo', 'test', curdir, 't.vtk'))


        # Create a dummy file in a subdir.
        s = BytesIO()
        # Spoof its location.
        s.name = abspath(join(cwd, 'data', 't.mv2'))
        # Dump into it.
        state_pickler.dump(t, s)

        # Rewind the stream
        s.seek(0)
        # "Move" the file elsewhere
        s.name = join(cwd, 'foo', 'test', 't.mv2')
        state = state_pickler.load_state(s)
        self.assertEqual(state.f.abs_pth,
                         join(cwd, 'foo', 't.vtk'))


if __name__ == "__main__":
    unittest.main()



import logging

from apptools.preferences.api import PreferencesHelper
from traits.api import Bool, Str, Trait


class LoggerPreferences(PreferencesHelper):
    """ The persistent service exposing the Logger plugin's API.
    """

    #### Preferences ###########################################################

    # The log levels
    level = Trait('Info',
        {'Debug'    : logging.DEBUG,
         'Info'     : logging.INFO,
         'Warning'  : logging.WARNING,
         'Error'    : logging.ERROR,
         'Critical' : logging.CRITICAL,
        },
        is_str = True,
    )

    enable_agent = Bool(False)
    smtp_server = Str()
    to_address = Str()
    from_address = Str()

    # The path to the preferences node that contains the preferences.
    preferences_path = Str('apptools.logger')



"""Tests for the version registry.

"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2005, Enthought, Inc.
# License: BSD Style.

# Standard library imports.
import sys
from imp import reload
import unittest

# Enthought library imports.
from traits.api import HasTraits

from apptools.persistence import version_registry, state_pickler


class Classic:
    __version__ = 0

class New(object):
    __version__ = 0

class TraitClass(HasTraits):
    __version__ = 0


class Test(New):
    __version__ = 1
    def __init__(self):
        self.a = Classic()

class Handler:
    def __init__(self):
        self.calls = []
    def upgrade(self, state, version):
        self.calls.append(('upgrade', state, version))
    def upgrade1(self, state, version):
        self.calls.append(('upgrade1', state, version))


class TestVersionRegistry(unittest.TestCase):
    def test_get_version(self):
        """Test the get_version function."""
        if sys.version_info[0] > 2:
            extra = [(('object', 'builtins'), -1)]
        else:
            extra = []
        c = Classic()
        v = version_registry.get_version(c)
        res = extra + [(('Classic', __name__), 0)]
        self.assertEqual(v, res)
        state = state_pickler.get_state(c)
        self.assertEqual(state.__metadata__['version'], res)

        n = New()
        v = version_registry.get_version(n)
        res = extra + [(('New', __name__), 0)]
        self.assertEqual(v, res)
        state = state_pickler.get_state(n)
        self.assertEqual(state.__metadata__['version'], res)

        t = TraitClass()
        v = version_registry.get_version(t)
        res = extra + [(('CHasTraits', 'traits.ctraits'), -1),
               (('HasTraits', 'traits.has_traits'), -1),
               (('TraitClass', __name__), 0)]
        self.assertEqual(v, res)
        state = state_pickler.get_state(t)
        self.assertEqual(state.__metadata__['version'], res)

    def test_reload(self):
        """Test if the registry is reload safe."""
        # A dummy handler.
        def h(x, y):
            pass

        registry = version_registry.registry
        registry.register('A', __name__, h)
        self.assertEqual(registry.handlers.get(('A', __name__)), h)
        reload(version_registry)
        registry = version_registry.registry
        self.assertEqual(registry.handlers.get(('A', __name__)), h)
        del registry.handlers[('A', __name__)]
        self.assertFalse(('A', __name__) in registry.handlers)

    def test_update(self):
        """Test if update method calls the handlers in order."""
        registry = version_registry.registry

        # First an elementary test.
        c = Classic()
        state = state_pickler.get_state(c)
        h = Handler()
        registry.register('Classic', __name__, h.upgrade)
        c1 = state_pickler.create_instance(state)
        state_pickler.set_state(c1, state)
        self.assertEqual(h.calls, [('upgrade', state, 0)])
        # Remove the handler.
        registry.unregister('Classic', __name__)

        # Now check to see if this works for inheritance trees.
        t = Test()
        state = state_pickler.get_state(t)
        h = Handler()
        registry.register('Classic', __name__, h.upgrade)
        registry.register('New', __name__, h.upgrade)
        registry.register('Test', __name__, h.upgrade1)
        t1 = state_pickler.create_instance(state)
        state_pickler.set_state(t1, state)
        # This should call New handler, then the Test and then
        # Classic.
        self.assertEqual(h.calls, [('upgrade', state, 0),
                                   ('upgrade1', state, 1),
                                   ('upgrade', state.a, 0)])



if __name__ == "__main__":
    unittest.main()



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" State factory for referenceable objects. """


# Local imports.
from .referenceable import Referenceable
from .state_factory import StateFactory


class ReferenceableStateFactory(StateFactory):
    """ State factory for referenceable objects. """

    ###########################################################################
    # 'StateFactory' interface.
    ###########################################################################

    def get_state_to_bind(self, obj, name, context):
        """ Returns the state of an object for binding. """

        state = None

        # If the object knows how to create a reference to it then let it
        # do so.
        if isinstance(obj, Referenceable):
            state = obj.reference

        return state

### EOF #######################################################################



def get_mro(obj_class):
    """ Get a reasonable method resolution order of a class and its
    superclasses for both old-style and new-style classes.
    """
    if not hasattr(obj_class, '__mro__'):
        # Old-style class. Mix in object to make a fake new-style class.
        try:
            obj_class = type(obj_class.__name__, (obj_class, object), {})
        except TypeError:
            # Old-style extension type that does not descend from object.
            mro = [obj_class]
        else:
            mro = obj_class.__mro__[1:-1]
    else:
        mro = obj_class.__mro__
    return mro


def _mod_name_key(typ):
    """ Return a '__module__:__name__' key for a type.
    """
    module = getattr(typ, '__module__', None)
    name = getattr(typ, '__name__', None)
    key = '{0}:{1}'.format(module, name)
    return key


class TypeRegistry(object):
    """ Register objects for types.

    Each type maintains a stack of registered objects that can be pushed and
    popped.
    """

    def __init__(self):
        # Map types to lists of registered objects. The last is the most
        # current and will be the one that is returned on lookup.
        self.type_map = {}

        # Map '__module__:__name__' strings to lists of registered objects.
        self.name_map = {}

        # Map abstract base classes to lists of registered objects.
        self.abc_map = {}

    #### TypeRegistry public interface ########################################

    def push(self, typ, obj):
        """ Push an object onto the stack for the given type.

        Parameters
        ----------
        typ : type or '__module__:__name__' string for a type
        obj : object
            The object to register.
        """
        if isinstance(typ, str):
            # Check the cached types.
            for cls in self.type_map:
                if _mod_name_key(cls) == typ:
                    self.type_map[cls].append(obj)
                    break
            else:
                if typ not in self.name_map:
                    self.name_map[typ] = []
                self.name_map[typ].append(obj)
        else:
            if typ not in self.type_map:
                self.type_map[typ] = []
            self.type_map[typ].append(obj)

    def push_abc(self, typ, obj):
        """ Push an object onto the stack for the given ABC.

        Parameters
        ----------
        typ : abc.ABCMeta
        obj : object
        """
        if typ not in self.abc_map:
            self.abc_map[typ] = []
        self.abc_map[typ].append(obj)

    def pop(self, typ):
        """ Pop a registered object for the given type.

        Parameters
        ----------
        typ : type or '__module__:__name__' string for a type

        Returns
        -------
        obj : object
            The last registered object for the type.

        Raises
        ------
        KeyError if the type is not registered.
        """
        if isinstance(typ, str):
            if typ not in self.name_map:
                # We may have it cached in the type map. We will have to
                # iterate over all of the types to check.
                for cls in self.type_map:
                    if _mod_name_key(cls) == typ:
                        old = self._pop_value(self.type_map, cls)
                        break
                else:
                    raise KeyError("No registered value for {0!r}".format(typ))
            else:
                old = self._pop_value(self.name_map, typ)
        else:
            if typ in self.type_map:
                old = self._pop_value(self.type_map, typ)
            elif typ in self.abc_map:
                old = self._pop_value(self.abc_map, typ)
            else:
                old = self._pop_value(self.name_map, _mod_name_key(typ))
        return old

    def lookup(self, instance):
        """ Look up the registered object for the given instance.

        Parameters
        ----------
        instance : object
            An instance of a possibly registered type.

        Returns
        -------
        obj : object
            The registered object for the type of the instance, one of the
            type's superclasses, or else one of the ABCs the type implements.

        Raises
        ------
        KeyError if the instance's type has not been registered.
        """
        return self.lookup_by_type(type(instance))

    def lookup_by_type(self, typ):
        """ Look up the registered object for a type.

        Parameters
        ----------
        typ  : type

        Returns
        -------
        obj : object
            The registered object for the type, one of its superclasses, or
            else one of the ABCs it implements.

        Raises
        ------
        KeyError if the type has not been registered.
        """
        return self.lookup_all_by_type(typ)[-1]

    def lookup_all(self, instance):
        """ Look up all the registered objects for the given instance.

        Parameters
        ----------
        instance : object
            An instance of a possibly registered type.

        Returns
        -------
        objs : list of objects
            The list of registered objects for the instance. If the given
            instance is not registered, its superclasses are searched. If none
            of the superclasses are registered, search the possible ABCs.

        Raises
        ------
        KeyError if the instance's type has not been registered.
        """
        return self.lookup_all_by_type(type(instance))

    def lookup_all_by_type(self, typ):
        """ Look up all the registered objects for a type.

        Parameters
        ----------
        typ  : type

        Returns
        -------
        objs : list of objects
            The list of registered objects for the type. If the given type is
            not registered, its superclasses are searched. If none of the
            superclasses are registered, search the possible ABCs.

        Raises
        ------
        KeyError if the type has not been registered.
        """
        # If a concrete superclass is registered use it.
        for cls in get_mro(typ):
            if cls in self.type_map or self._in_name_map(cls):
                objs = self.type_map[cls]
                if objs:
                    return objs

        # None of the concrete superclasses. Check the ABCs.
        for abstract, objs in self.abc_map.items():
            if issubclass(typ, abstract) and objs:
                return objs

        # If we have reached here, the lookup failed.
        raise KeyError("No registered value for {0!r}".format(typ))

    #### Private implementation ###############################################

    def _pop_value(self, mapping, key):
        """ Pop a value from a keyed stack in a mapping, taking care to remove
        the key if the stack is depleted.
        """
        objs = mapping[key]
        old = objs.pop()
        if not objs:
            del mapping[key]
        return old

    def _in_name_map(self, typ):
        """ Check if the given type is specified in the name map.

        Parameters
        ----------
        typ : type

        Returns
        -------
        is_in_name_map : bool
            If True, the registered value will be moved over to the type map
            for future lookups.
        """
        key = _mod_name_key(typ)
        if key in self.name_map:
            self.type_map[typ] = self.name_map.pop(key)
            return True
        else:
            return False


class LazyRegistry(TypeRegistry):
    """ A type registry that will lazily import the registered objects.

    Register '__module__:__name__' strings for the lazily imported objects.
    These will only be imported when the matching type is looked up. The module
    name must be a fully-qualified absolute name with all of the parent
    packages specified.
    """

    def lookup_by_type(self, typ):
        """ Look up the registered object for a type.
        """
        mod_name = TypeRegistry.lookup_by_type(self, typ)
        return self._import_object(mod_name)

    def _import_object(self, mod_object):
        module, name = mod_object.split(':')
        mod = __import__(module, {}, {}, [name], 0)
        return getattr(mod, name)



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" A naming context for a Python namespace. """


# Enthought library imports.
from traits.api import Any, Dict, Instance, Property

# Local imports.
from .address import Address
from .binding import Binding
from .context import Context
from .naming_manager import naming_manager
from .py_object_factory import PyObjectFactory
from .reference import Reference
from .referenceable import Referenceable
from .referenceable_state_factory import ReferenceableStateFactory


# The default environment.
ENVIRONMENT = {
    # 'Context' properties.
    Context.OBJECT_FACTORIES : [PyObjectFactory()],
    Context.STATE_FACTORIES  : [ReferenceableStateFactory()],
}


class PyContext(Context, Referenceable):
    """ A naming context for a Python namespace. """

    #### 'Context' interface ##################################################

    # The naming environment in effect for this context.
    environment = Dict(ENVIRONMENT)

    #### 'PyContext' interface ################################################

    # The Python namespace that we represent.
    namespace = Any

    # If the namespace is actual a Python object that has a '__dict__'
    # attribute, then this will be that object (the namespace will be the
    # object's '__dict__'.
    obj = Any

    #### 'Referenceable' interface ############################################

    # The object's reference suitable for binding in a naming context.
    reference = Property(Instance(Reference))

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """ Creates a new context. """

        # Base class constructor.
        super(PyContext, self).__init__(**traits)

        if type(self.namespace) is not dict:
            if hasattr(self.namespace, '__dict__'):
                self.obj = self.namespace
                self.namespace = self.namespace.__dict__

            else:
                raise ValueError('Need a dictionary or a __dict__ attribute')

        return

    ###########################################################################
    # 'Referenceable' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_reference(self):
        """ Returns a reference to this object suitable for binding. """

        reference = Reference(
            class_name = self.__class__.__name__,
            addresses  = [Address(type='py_context', content=self.namespace)]
        )

        return reference

    ###########################################################################
    # Protected 'Context' interface.
    ###########################################################################

    def _is_bound(self, name):
        """ Is a name bound in this context? """

        return name in self.namespace

    def _lookup(self, name):
        """ Looks up a name in this context. """

        obj = self.namespace[name]

        return naming_manager.get_object_instance(obj, name, self)

    def _bind(self, name, obj):
        """ Binds a name to an object in this context. """

        state = naming_manager.get_state_to_bind(obj, name,self)
        self.namespace[name] = state

        # Trait event notification.
        # An "added" event is fired by the bind method of the base calss (which calls
        # this one), so we don't need to do the changed here (which would be the wrong
        # thing anyway) -- LGV
        #
        # self.trait_property_changed('context_changed', None, None)

        return

    def _rebind(self, name, obj):
        """ Rebinds a name to a object in this context. """

        self._bind(name, obj)

        return

    def _unbind(self, name):
        """ Unbinds a name from this context. """

        del self.namespace[name]

        # Trait event notification.
        self.trait_property_changed('context_changed', None, None)

        return

    def _rename(self, old_name, new_name):
        """ Renames an object in this context. """

        state = self.namespace[old_name]

        # Bind the new name.
        self.namespace[new_name] = state

        # Unbind the old one.
        del self.namespace[old_name]

        # Trait event notification.
        self.context_changed = True

        return

    def _create_subcontext(self, name):
        """ Creates a sub-context of this context. """

        sub = self._context_factory(name, {})
        self.namespace[name] = sub

        # Trait event notification.
        self.trait_property_changed('context_changed', None, None)

        return sub

    def _destroy_subcontext(self, name):
        """ Destroys a sub-context of this context. """

        del self.namespace[name]

        # Trait event notification.
        self.trait_property_changed('context_changed', None, None)

        return

    def _list_bindings(self):
        """ Lists the bindings in this context. """

        bindings = []
        for name, value in list(self.namespace.items()):
            bindings.append(Binding(name=name, obj=self._lookup(name),
                                    context=self))
        return bindings

    def _list_names(self):
        """ Lists the names bound in this context. """

        return list(self.namespace.keys())

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _context_factory(self, name, namespace):
        """ Create a sub-context. """

        return self.__class__(namespace=namespace)

#### EOF ######################################################################




# Standard library imports
import sys
import pickle
import logging

# Enthought library imports
from apptools.persistence.versioned_unpickler import VersionedUnpickler


logger = logging.getLogger(__name__)


def load_project(pickle_filename, updater_path, application_version, protocol,
                                                                  max_pass=-1):
    """ Reads a project from a pickle file and if necessary will update it to
    the latest version of the application.
    """

    latest_file = pickle_filename

    # Read the pickled project's metadata.
    f = open(latest_file, 'rb')
    metadata = VersionedUnpickler(f).load(max_pass)
    f.close()
    project_version = metadata.get('version', False)

    if not project_version:
        raise ValueError("Could not read version number from the project file")

    logger.debug('Project version: %d, Application version: %d' %
                (project_version, application_version))

    # here you can temporarily force an upgrade each time for testing ....
    # project_version = 0
    latest_file = upgrade_project(pickle_filename, updater_path,
                                project_version, application_version, protocol,
                                max_pass)

    # Finally we can import the project ...
    logger.info('loading %s' % latest_file)
    i_f = open(latest_file, 'rb')
    version = VersionedUnpickler(i_f).load(max_pass)
    project = VersionedUnpickler(i_f).load(max_pass)
    i_f.close()

    return project


def upgrade_project(pickle_filename, updater_path, project_version, application_version, protocol, max_pass=-1):
    """ Repeatedly read and write the project to disk updating it one version
    at a time.

    Example the p5.project is at version 0
    The application is at version 3

    p5.project    --- Update1 ---> p5.project.v1
    p5.project.v1 --- Update2 ---> p5.project.v2
    p5.project.v2 --- Update3 ---> p5.project.v3
    p5.project.v3 ---> loaded into app

    The user then has the option to save the updated project as p5.project
    """
    first_time = True
    latest_file = pickle_filename

    # update the project until it's version matches the application's
    while project_version < application_version:

        next_version = project_version + 1

        if first_time:
            i_f = open(pickle_filename, 'rb')
            data = i_f.read()
            open('%s.bak' % pickle_filename, 'wb').write(data)
            i_f.seek(0) # rewind the file to the start
        else:
            name = '%s.v%d' % (pickle_filename, project_version)
            i_f = open(name, 'rb')
            latest_file = name

        logger.info('converting %s' % latest_file)

        # find this version's updater ...
        updater_name = '%s.update%d' % (updater_path, next_version)
        __import__(updater_name)
        mod = sys.modules[updater_name]
        klass = getattr(mod, 'Update%d' % next_version)
        updater = klass()

        # load and update this version of the project
        version = VersionedUnpickler(i_f).load(max_pass)
        project = VersionedUnpickler(i_f, updater).load(max_pass)
        i_f.close()

        # set the project version to be the same as the updater we just
        # ran on the unpickled files ...
        project.metadata['version'] = next_version

        # Persist the updated project ...
        name = '%s.v%d' % (pickle_filename, next_version)
        latest_file = name
        o_f = open(name, 'wb')
        pickle.dump(project.metadata, o_f, protocol=protocol)
        pickle.dump(project, o_f, protocol=protocol)
        o_f.close()

        # Bump up the version number of the pickled project...
        project_version += 1
        first_time = False

    return latest_file


### EOF #################################################################




""" A default adapter class. """


# Enthought library imports.
from traits.api import Any, HasTraits


class Adapter(HasTraits):
    """ A default adapter class.

    This comes in handy when using the default 'AdapterFactory' as it expects
    the adpter class to have an 'adaptee' trait.

    """

    #### 'Adapter' interface ##################################################

    # The object that we are adapting.
    adaptee = Any

#### EOF ######################################################################



""" An object that can be initialized from a preferences node. """


# Standard library imports.
import logging

# Enthought library imports.
from traits.api import HasTraits, Instance, Str, Unicode

# Local imports.
from .i_preferences import IPreferences
from .package_globals import get_default_preferences


# Logging.
logger = logging.getLogger(__name__)


class PreferencesHelper(HasTraits):
    """ An object that can be initialized from a preferences node. """

    #### 'PreferencesHelper' interface ########################################

    # The preferences node used by the helper. If this trait is not set then
    # the package-global default preferences node is used.
    #
    # fixme: This introduces a 'sneaky' global reference to the preferences
    # node!
    preferences = Instance(IPreferences)

    # The path to the preference node that contains the preferences that we
    # use to initialize instances of this class.
    preferences_path = Str

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """ Constructor. """

        super(PreferencesHelper, self).__init__(**traits)

        # Initialize the object's traits from the preferences node.
        if self.preferences:
            self._initialize(self.preferences)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _preferences_default(self):
        """ Trait initializer. """

        # If no specific preferences node is set then we use the package-wide
        # global node.
        return get_default_preferences()

    #### Trait change handlers ################################################

    def _anytrait_changed(self, trait_name, old, new):
        """ Static trait change handler. """

        # If we were the one that set the trait (because the underlying
        # preferences node changed) then do nothing.
        if self.preferences and self._is_preference_trait(trait_name):
            self.preferences.set('%s.%s' % (self._get_path(), trait_name), new)

        return

    def _preferences_changed(self, old, new):
        """ Static trait change handler. """

        # Stop listening to the old preferences node.
        if old is not None:
            old.remove_preferences_listener(
                self._preferences_changed_listener, self._get_path()
            )

        if new is not None:
            # Initialize with the new preferences node (this also adds a
            # listener for preferences being changed in the new node).
            self._initialize(new, notify=True)

        return

    #### Other observer pattern listeners #####################################

    def _preferences_changed_listener(self, node, key, old, new):
        """ Listener called when a preference value is changed. """

        if key in self.trait_names():
            setattr(self, key, self._get_value(key, new))

        return

    #### Methods ##############################################################

    def _get_path(self):
        """ Return the path to our preferences node. """

        if len(self.preferences_path) > 0:
            path = self.preferences_path

        else:
            path = getattr(self, 'PREFERENCES_PATH', None)
            if path is None:
                raise SystemError('no preferences path, %s' % self)

            else:
                logger.warn('DEPRECATED: use "preferences_path" %s' % self)

        return path

    def _get_value(self, trait_name, value):
        """ Get the actual value to set.

        This method makes sure that any required work is done to convert the
        preference value from a string. Str traits or those with the metadata
        'is_str=True' will just be passed the string itself.

        """

        trait = self.trait(trait_name)
        handler = trait.handler

        # If the trait type is 'Str' or Unicode then we just take the raw value.
        if isinstance(handler, (Str, Unicode)) or trait.is_str:
            pass

        # Otherwise, we eval it!
        else:
            try:
                value = eval(value)

            # If the eval fails then there is probably a syntax error, but
            # we will let the handler validation throw the exception.
            except:
                pass

        if handler.validate is not None:
            # Any traits have a validator of None.
            validated = handler.validate(self, trait_name, value)
        else:
            validated = value

        return validated

    def _initialize(self, preferences, notify=False):
        """ Initialize the object's traits from the preferences node. """

        path = self._get_path()
        keys = preferences.keys(path)

        traits_to_set = {}
        for trait_name in self.trait_names():
            if trait_name in keys:
                key = '%s.%s' % (path, trait_name)
                value = self._get_value(trait_name, preferences.get(key))
                traits_to_set[trait_name] = value

        self.set(trait_change_notify=notify, **traits_to_set)

        # Listen for changes to the node's preferences.
        preferences.add_preferences_listener(
            self._preferences_changed_listener, path
        )

        return

    # fixme: Pretty much duplicated in 'PreferencesPage' (except for the
    # class name of course!).
    def _is_preference_trait(self, trait_name):
        """ Return True if a trait represents a preference value. """

        if trait_name.startswith('_') or trait_name.endswith('_') \
           or trait_name in PreferencesHelper.class_traits():
            return False

        return True

#### EOF ######################################################################



"""
A Recorder subclass that presents a simple user interface.
"""
# Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
# Copyright (c) 2008, Prabhu Ramachandran.
# License: BSD Style.

from traits.api import Code, Button, Int, on_trait_change, Any
from traitsui.api import (View, Item, Group, HGroup, CodeEditor,
                                     spring, Handler)

from .recorder import Recorder

######################################################################
# `CloseHandler` class.
######################################################################
class CloseHandler(Handler):
    """This class cleans up after the UI for the recorder is closed."""

    def close(self, info, is_ok):
        """This method is invoked when the user closes the UI."""
        recorder = info.object
        recorder.on_ui_close()
        return True


################################################################################
# `RecorderWithUI` class.
################################################################################
class RecorderWithUI(Recorder):
    """
    This class represents a Recorder but with a simple user interface.
    """

    # The code to display
    code = Code(editor=CodeEditor(line='current_line'))

    # Button to save script to file.
    save_script = Button('Save Script')

    # The current line to show, used by the editor.
    current_line = Int

    # The root object which is being recorded.
    root = Any

    ########################################
    # Traits View.
    view = View(
             Group(
                HGroup(Item('recording', show_label=True),
                       spring,
                       Item('save_script', show_label=False),
                ),
                Group(Item('code', show_label=False)),
                ),
             width=600,
             height=360,
             id='apptools.scripting.recorder_with_ui',
             buttons=['Cancel'],
             resizable=True,
             handler=CloseHandler()
             )

    ######################################################################
    # RecorderWithUI interface.
    ######################################################################
    def on_ui_close(self):
        """Called from the CloseHandler when the UI is closed. This
        method basically stops the recording.
        """
        from .util import stop_recording
        from .package_globals import get_recorder

        if get_recorder() is self:
            stop_recording(self.root, save=False)
        else:
            self.recording = False
            self.unregister(self.root)

    ######################################################################
    # Non-public interface.
    ######################################################################
    @on_trait_change('lines[]')
    def _update_code(self):
        self.code = self.get_code()
        self.current_line = len(self.lines) + 1

    def _save_script_fired(self):
        self.ui_save()



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from pyface.action.api import Group, MenuManager
from traits.api import Unicode

# Local imports.
from apptools.permissions.package_globals import get_permissions_manager
from .login_action import LoginAction
from .logout_action import LogoutAction


class UserMenuManager(MenuManager):
    """A menu that contains all the actions related to users and permissions.
    """

    #### 'MenuManager' interface ##############################################

    id = 'User'

    name = Unicode("&User")

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """Initialise the object."""

        pm = get_permissions_manager()

        # Put them in a group so we can optionally append (because the PyFace
        # API doesn't do what you expect with append()).
        group = Group()

        group.append(LoginAction())

        for act in pm.user_manager.user_actions:
            group.append(act)

        group.append(LogoutAction())

        for act in pm.user_manager.management_actions:
            group.append(act)

        for act in pm.policy_manager.management_actions:
            group.append(act)

        super(UserMenuManager, self).__init__(group, **traits)



import abc


class A(object):
    pass


class B(A):
    pass


class C(B):
    pass


class D(object):
    pass


class Mixed(A, D):
    pass


class Abstract(object, metaclass=abc.ABCMeta):
    pass


class Concrete(object):
    pass


class ConcreteSubclass(Concrete):
    pass


for typ in (A, B, C, D, Mixed, Abstract, Concrete, ConcreteSubclass):
    typ.__module__ = 'dummies'


Abstract.register(Concrete)



#-------------------------------------------------------------------------------
#
#  Defines the ITemplateDataSource interface for creating 'live' application
#  data sources from a templatized data source object.
#
#  Written by: David C. Morrill
#
#  Date: 07/26/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" Defines the ITemplateDataSource interface for creating 'live' application
    data sources from a templatized data source object.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import Interface

#-------------------------------------------------------------------------------
#  'ITemplateDataSource' interface:
#-------------------------------------------------------------------------------

class ITemplateDataSource ( Interface ):
    """ Defines the ITemplateDataSource interface for creating 'live'
        application data sources from a templatized data source object.
    """

    def name_from_data_source ( self ):
        """ Allows the object to provide a description of the possibly optional
            data binding it requires.

            Returns
            -------
            A **TemplateDataName** object describing the binding the data
            source object requires.
        """




#-------------------------------------------------------------------------------
#
#  Defines the ITemplateDataNameItem interface used by elements of a
#  TemplateDataName.
#
#  Written by: David C. Morrill
#
#  Date: 07/27/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" Defines the ITemplateDataNameItem interface used by elements of a
    TemplateDataName.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import Interface, Instance, List

from .itemplate_data_context \
    import ITemplateDataContext

from .itemplate_choice \
    import ITemplateChoice

#-------------------------------------------------------------------------------
#  'ITemplateDataNameItem' interface:
#-------------------------------------------------------------------------------

class ITemplateDataNameItem ( Interface ):
    """ Defines the ITemplateDataNameItem interface used by elements of a
        TemplateDataName.

        The contents of the *input_data_context* and *output_data_context*
        are assumed to be immutable. That is, no changes should occur to the
        data contexts once they have been assigned to the
        **ITemplateDataNameItem** traits.

        However, new input contexts can be assigned, in which case a new
        output context should be created and assigned to the output context
        (once it has been fully populated).

        Also, the *output_data_context* should be **None** if the
        *input_data_context* value is **None** or the object cannot match any
        values in the *input_data_context*.
    """

    # The data context which this data name item should match against. This
    # value must be read/write.
    input_data_context = Instance( ITemplateDataContext )

    # The data context containing the data values and/or contexts this data
    # name item matches. This value must be read/write.
    output_data_context = Instance( ITemplateDataContext )

    # The ITemplateChoice instance representing the current settings of the
    # data name item. This value must be read/write.
    data_name_item_choice = Instance( ITemplateChoice )

    # The alternative choices the user has for the data name item settings for
    # the current input data context. The list may be empty, in which case the
    # user cannot change the settings of the data name item. This value can be
    # read only.
    data_name_item_choices = List( ITemplateChoice )





""" Context adapter for Python lists. """


# Enthought library imports.
from apptools.naming.api import Binding, ContextAdapter, naming_manager
from traits.api import List, Property


class ListContextAdapter(ContextAdapter):
    """ Context adapter for Python lists. """

    #### 'ContextAdapter' interface ###########################################

    # The object that we are adapting.
    #
    # fixme: We would like to specialize the 'adaptee' trait here, but if we
    # make it of type 'List' then, on assignment, traits will create a *copy*
    # of the actual list which I think you'll agree is not very adapter-like!
##     adaptee = List

    #### 'ListContextAdapter' interface #######################################

    # The list that we are adapting.
    collection = Property(List)

    ###########################################################################
    # Protected 'Context' interface.
    ###########################################################################

    def _is_bound(self, name):
        """ Is a name bound in this context? """

        return name in self._list_names()

    def _lookup(self, name):
        """ Looks up a name in this context. """

        binding = self._get_binding_with_name(name)

        return naming_manager.get_object_instance(binding.obj, name, self)

    def _bind(self, name, obj):
        """ Binds a name to an object in this context. """

        state = naming_manager.get_state_to_bind(obj, name, self)
        self.collection.append(state)

        return

    def _rebind(self, name, obj):
        """ Rebinds a name to an object in this context. """

        index = 0
        for binding in self.list_bindings(''):
            if binding.name == name:
                self.collection[index] = obj
                break

            index = index + 1

        # The name is not already bound.
        else:
            self._bind(name, obj)

        return

    def _unbind(self, name):
        """ Unbinds a name from this context. """

        index = 0
        for binding in self.list_bindings(''):
            if binding.name == name:
                del self.collection[index]
                break

            index = index + 1

        else:
            raise SystemError('no binding with name %s' % name)

        return

    def _rename(self, old_name, new_name):
        """ Renames an object in this context. """

        binding = self._get_binding_with_name(old_name)

        self._set_name(binding.obj, new_name)

        return

    def _create_subcontext(self, name):
        """ Creates a sub-context of this context. """

        raise OperationNotSupportedError()

    def _destroy_subcontext(self, name):
        """ Destroys a sub-context of this context. """

        raise OperationNotSupportedError()

    def _list_bindings(self):
        """ Lists the bindings in this context. """

        bindings = []

        for obj in self.collection:
            # Bindings have to have a string name.
            name = self._get_name(obj)

            # Create the binding.
            bindings.append(Binding(name=name, obj=obj, context=self))

        return bindings

    def _list_names(self):
        """ Lists the names bound in this context. """

        return [self._get_name(obj) for obj in self.collection]

    ###########################################################################
    # Protected 'ListContext' interface.
    ###########################################################################

    def _get_collection(self):
        """ Returns the collection that we are adapting. """

        return self.adaptee

    ###########################################################################
    # Private interface.
    ###########################################################################

    # fixme: Allow an item name trait to be specified instead of guessing at
    # 'name' or 'id'!
    def _get_name(self, obj):
        """ Returns the name of an object. """

        if hasattr(obj, 'name'):
            name = str(obj.name)

        elif hasattr(obj, 'id'):
            name = str(obj.id)

        else:
            name = str(obj)

        return name

    def _set_name(self, obj, name):
        """ Sets the name of an object. """

        if hasattr(obj, 'name'):
            obj.name = name

        elif hasattr(obj, 'id'):
            obj.id = name

        return

    def _get_binding_with_name(self, name):
        """ Returns the binding with the specified name. """

        for binding in self.list_bindings(''):
            if binding.name == name:
                break

        # The reason that this is a system error and not just a naming error
        # is that this method is only called from inside the protected
        # 'Context' interface when we have already determined that the name
        # is bound
        else:
            raise SystemError('no binding with name %s' % name)

        return binding

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Interface


class PolicyStorageError(Exception):
    """This is the exception raised by an IPolicyStorage object when an error
    occurs accessing the database.  Its string representation is displayed as
    an error message to the user."""


class IPolicyStorage(Interface):
    """This defines the interface expected by a PolicyManager to handle the low
    level storage of the policy data."""

    ###########################################################################
    # 'IPolicyStorage' interface.
    ###########################################################################

    def add_role(self, name, description, perm_ids):
        """Add a new role."""

    def all_roles(self):
        """Return a list of all roles where each element is a tuple of the name
        and description."""

    def delete_role(self, name):
        """Delete the role with the given name (which will not be empty)."""

    def get_assignment(self, user_name):
        """Return a tuple of the user name and list of role names of the
        assignment for the given user name.  The tuple will contain an empty
        string and list if the user isn't known."""

    def get_policy(self, user_name):
        """Return a tuple of the user name and list of permission names for the
        user with the given name.  The tuple will contain an empty string and
        list if the user isn't known."""

    def is_empty(self):
        """Return True if the user database is empty.  It will only ever be
        called once."""

    def matching_roles(self, name):
        """Return a list of tuples of the full name, description and list of
        permission names, sorted by the full name, of all roles that match the
        given name.  How the name is interpreted (eg. as a regular expression)
        is determined by the storage."""

    def modify_role(self, name, description, perm_ids):
        """Update the description and permissions for the role with the given
        name (which will not be empty)."""

    def set_assignment(self, user_name, role_names):
        """Save the assignment of the given role names to the given user.  Note
        that there may or may not be an existing assignment for the user."""



""" Context adapter for trait dictionaries. """


# Enthought library imports.
from traits.api import Dict, Property, Str

# Local imports.
from .dict_context_adapter import DictContextAdapter


class TraitDictContextAdapter(DictContextAdapter):
    """ Context adapter for trait dictionaries. """

    #### 'Context' interface ##################################################

    # The name of the context within its own namespace.
    namespace_name = Property(Str)

    #### 'ContextAdapter' interface ###########################################

    # The object that we are adapting.
    adaptee = Dict

    #### 'TraitDictContextAdapter' interface ##################################

    # The name of the object's trait that provides the dictionary.
    trait_name = Str

    ###########################################################################
    # 'Context' interface.
    ###########################################################################

    def _get_namespace_name(self):
        """ Returns the name of the context within its own namespace. """

        return self.context.namespace_name + '/' + self.trait_name

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought logger package component>
#------------------------------------------------------------------------------
""" Prints a stack trace every time it is called but does not halt execution
    of the application.

    Copied from Uche Ogbuji's blog
"""

# Standard library imports.
import inspect
from io import StringIO


def log_point(msg='\n'):
    stack = inspect.stack()
    # get rid of logPoint's part of the stack:
    stack = stack[1:]
    stack.reverse()
    output = StringIO()
    if msg:
        output.write(str(msg) + '\n')
    for stackLine in stack:
        frame, filename, line, funcname, lines, unknown = stackLine
        if filename.endswith('/unittest.py'):
            # unittest.py code is a boring part of the traceback
            continue
        if filename.startswith('./'):
            filename = filename[2:]
        output.write('%s:%s in %s:\n' % (filename, line, funcname))
        if lines:
            output.write('  %s\n' % ''.join(lines)[:-1])
    s = output.getvalue()

    return s

## EOF ##################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" Object factory for Python namespace contexts. """


# Local imports.
from .object_factory import ObjectFactory
from .reference import Reference


class PyObjectFactory(ObjectFactory):
    """ Object factory for Python namespace contexts. """

    ###########################################################################
    # 'ObjectFactory' interface.
    ###########################################################################

    def get_object_instance(self, state, name, context):
        """ Creates an object using the specified state information. """

        obj = None

        if isinstance(state, Reference):
            if len(state.addresses) > 0:
                if state.addresses[0].type == 'py_context':
                    namespace = state.addresses[0].content
                    obj = context._context_factory(name, namespace)

        elif hasattr(state, '__dict__'):
            from apptools.naming.py_context import PyContext
            if not isinstance(state, PyContext):
                obj = context._context_factory(name, state)

        return obj

### EOF #######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" A tree view of a naming system. """


# Enthought library imports.
from apptools.naming.api import OperationNotSupportedError
from pyface.tree.api import NodeTree
from traits.api import Instance

# Local imports.
from .naming_tree_model import NamingTreeModel


class NamingTree(NodeTree):
    """ A tree view of a naming system. """

    #### 'Tree' interface #####################################################

    # The model that provides the data for the tree.
    model = Instance(NamingTreeModel)

    ###########################################################################
    # 'Tree' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _model_default(self):
        """ Initializes the model trait. """

        return NamingTreeModel()

    ###########################################################################
    # 'NamingTree' interface.
    ###########################################################################

    def ensure_visible(self, node):
        """ Make sure that the specified node is visible. """

        try:
            components = node.namespace_name.split('/')

            # Make sure that the tree is expanded down to the context that
            # contains the node.
            binding = self.root
            for atom in components[:-1]:
                binding = binding.obj.lookup_binding(atom)
                self.expand(binding)

            # The context is expanded so we know that the node will be in the
            # node to Id map.
            wxid = self._node_to_id_map.get(self.model.get_key(node), None)
            self.control.EnsureVisible(wxid)

        # We need 'namespace_name' to make this work.  If we don't have it
        # then we simply cannot do this!
        except OperationNotSupportedError:
            binding = None

        return binding

##### EOF #####################################################################



""" (Pyface) Action for running a help demo.

    :Copyright: 2008, Enthought Inc.
    :License: BSD
    :Author: Janet Swisher
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!

# Standard library imports.
import logging
from os.path import isabs, join, normpath
from subprocess import Popen
import sys

# Enthought library imports.
from envisage.api import IExtensionPointUser, IExtensionRegistry
from pyface.workbench.action.workbench_action import WorkbenchAction
from traits.api import Instance, provides, Property
from pyface.api import ImageResource
from apptools.help.help_plugin.api import HelpCode

# Logging.
logger = logging.getLogger(__name__)

# This module's parent package.
PARENT = '.'.join(__name__.split('.')[:-2])

from .util import get_sys_prefix_relative_filename

# Implementation of the ImageResource class to be used for the DocAction class.
@provides(IExtensionPointUser)
class DemoImageResource(ImageResource):
    """ Implementation of the ImageResource class to be used for the DemoAction
    class.
    Overrides the '_image_not_found' trait in the base ImageResource class.
    """

    # Image to display when the specified image file cannot be located.
    _image_not_found = ImageResource('python_run')

class DemoAction(WorkbenchAction):
    """
    """


    ### Action interface ##############################################

    # Image associated with this action instance.
    image = Property

    ### IExtensionPointUser interface
    extension_registry = Property(Instance(IExtensionRegistry))

    def _get_extension_registry(self):
        return self.window.application.extension_registry

    def _get_image(self):
        """ Returns the image to be used for this DemoAction instance.
        """
        # The current implementation searches for an image file matching
        # 'name' in all of the image paths. If such a file is not to be found,
        # the '_image_not_found' file for the DemoImageResourceClass is used.
        return DemoImageResource(self.name)

    ### HelpDemoAction interface

    # Help code associated with this action.
    my_help_code = Instance(HelpCode)

    def _my_help_code_default(self):
        exns = self.extension_registry.get_extensions(PARENT + '.help_demos')
        for hc in exns:
            if hc.label == self.name:
                return hc
        return None

    def perform(self, event):
        """ Perform the action by running the demo.
        """
        if self.my_help_code is not None:
            if self.my_help_code.filename:
                filename = get_sys_prefix_relative_filename(self.my_help_code.filename)
                if filename is not None:
                    try:
                        Popen([sys.executable, filename])
                    except OSError as err:
                        logger.error(
                                'Could not execute Python file for Demo "%s".\n\n' \
                                 % self.my_help_code.label + str(err) + \
                                 '\n\nTry changing Demo Preferences.' )
            elif self.my_help_code.code:
                exec("%s" % self.my_help_code.code)
        return



""" An adapter factory. """


# Enthought library imports.
from traits.api import Any

# Local imports.
from .abstract_adapter_factory import AbstractAdapterFactory


class AdapterFactory(AbstractAdapterFactory):
    """ An adapter factory.

    This implementation provides the common case where the factory adapts
    exactly one type of object to exactly one target type using exactly one
    adapter.

    This class requires the adapter class to have an 'adaptee' trait. The
    default 'Adapter' class provides exactly that.

    """

    #### 'AdapterFactory' interface ###########################################

    # fixme: These trait definitions should be 'Class' but currently this only
    # allows old-style classes!
    #
    # The type of object that the factory can adapt.
    adaptee_class = Any

    # The adapter class (the class that adapts the adaptee to the target
    # class).
    adapter_class = Any

    # The target class (the class that the factory can adapt objects to).
    target_class = Any

    ###########################################################################
    # Protected 'AbstractAdapterFactory' interface.
    ###########################################################################

    def _can_adapt(self, adaptee, target_class, *args, **kw):
        """ Returns True if the factory can produce an appropriate adapter. """

        can_adapt = target_class is self.target_class \
                    and self.type_system.is_a(adaptee, self.adaptee_class)

        return can_adapt

    def _adapt(self, adaptee, target_class, *args, **kw):
        """ Returns an adapter that adapts an object to the target class.

        This requires the adapter class to have an 'adaptee' trait. The default
        'Adapter' class provides exactly that.

        """

        return self.adapter_class(adaptee=adaptee, *args, **kw)

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# The permissions manager.
_permissions_manager = None


def get_permissions_manager():
    """Return the IPermissionsManager implementation, creating a
    PermissionsManager instance if no other implementation has been set."""

    global _permissions_manager

    if _permissions_manager is None:
        from .permissions_manager import PermissionsManager

        _permissions_manager = PermissionsManager()

    return _permissions_manager


def set_permissions_manager(permissions_manager):
    """Set the IPermissionsManager implementation to use."""

    global _permissions_manager

    _permissions_manager = permissions_manager



#------------------------------------------------------------------------------
# Copyright (c) 2006 by Enthought, Inc.
# All rights reserved.
#------------------------------------------------------------------------------
""" Manages type extensions, including factories to generate adapters, and
    hooks for methods and functions. Part of the AppTools project of the
    Enthought Tool Suite.
"""
from .api import *



""" Tests for preference bindings. """


# Standard library imports.
import os, tempfile, unittest
from os.path import join

# Major package imports.
from pkg_resources import resource_filename

# Enthought library imports.
from apptools.preferences.api import Preferences, PreferenceBinding
from apptools.preferences.api import ScopedPreferences, bind_preference
from apptools.preferences.api import set_default_preferences
from traits.api import Bool, HasTraits, Int, Float, Str


# This module's package.
PKG = 'apptools.preferences.tests'


def listener(obj, trait_name, old, new):
    """ A useful trait change handler for testing! """

    listener.obj = obj
    listener.trait_name = trait_name
    listener.old = old
    listener.new = new

    return


class PreferenceBindingTestCase(unittest.TestCase):
    """ Tests for preference bindings. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        self.preferences = set_default_preferences(Preferences())

        # The filename of the example preferences file.
        self.example = resource_filename(PKG, 'example.ini')

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_preference_binding(self):
        """ preference binding """

        p = self.preferences
        p.load(self.example)

        class AcmeUI(HasTraits):
            """ The Acme UI class! """

            # The traits that we want to initialize from preferences.
            bgcolor = Str
            width   = Int
            ratio   = Float
            visible = Bool

        acme_ui = AcmeUI()
        acme_ui.on_trait_change(listener)

        # Make some bindings.
        bind_preference(acme_ui, 'bgcolor', 'acme.ui.bgcolor')
        bind_preference(acme_ui, 'width',   'acme.ui.width')
        bind_preference(acme_ui, 'ratio',   'acme.ui.ratio')
        bind_preference(acme_ui, 'visible', 'acme.ui.visible')

        # Make sure the object was initialized properly.
        self.assertEqual('blue', acme_ui.bgcolor)
        self.assertEqual(50, acme_ui.width)
        self.assertEqual(1.0, acme_ui.ratio)
        self.assertEqual(True, acme_ui.visible)

        # Make sure we can set the preference via the helper...
        acme_ui.bgcolor = 'yellow'
        self.assertEqual('yellow', p.get('acme.ui.bgcolor'))
        self.assertEqual('yellow', acme_ui.bgcolor)

        # ... and that the correct trait change event was fired.
        self.assertEqual(acme_ui, listener.obj)
        self.assertEqual('bgcolor', listener.trait_name)
        self.assertEqual('blue', listener.old)
        self.assertEqual('yellow', listener.new)

        # Make sure we can set the preference via the preferences node...
        p.set('acme.ui.bgcolor', 'red')
        self.assertEqual('red', p.get('acme.ui.bgcolor'))
        self.assertEqual('red', acme_ui.bgcolor)

        # ... and that the correct trait change event was fired.
        self.assertEqual(acme_ui, listener.obj)
        self.assertEqual('bgcolor', listener.trait_name)
        self.assertEqual('yellow', listener.old)
        self.assertEqual('red', listener.new)

        # Make sure we can set a non-string preference via the helper...
        acme_ui.ratio = 0.5
        self.assertEqual('0.5', p.get('acme.ui.ratio'))
        self.assertEqual(0.5, acme_ui.ratio)

        # Make sure we can set a non-string preference via the node...
        p.set('acme.ui.ratio', '0.75')
        self.assertEqual('0.75', p.get('acme.ui.ratio'))
        self.assertEqual(0.75, acme_ui.ratio)

        return

    def test_default_values(self):
        """ instance scope preferences path """

        p = self.preferences

        class AcmeUI(HasTraits):
            """ The Acme UI class! """

            # The traits that we want to initialize from preferences.
            bgcolor = Str('blue')
            width   = Int(50)
            ratio   = Float(1.0)
            visible = Bool(True)

        acme_ui = AcmeUI()

        # Make some bindings.
        bind_preference(acme_ui, 'bgcolor', 'acme.ui.bgcolor')
        bind_preference(acme_ui, 'width',   'acme.ui.width')
        bind_preference(acme_ui, 'ratio',   'acme.ui.ratio')
        bind_preference(acme_ui, 'visible', 'acme.ui.visible')

        # Make sure the helper was initialized properly.
        self.assertEqual('blue', acme_ui.bgcolor)
        self.assertEqual(50, acme_ui.width)
        self.assertEqual(1.0, acme_ui.ratio)
        self.assertEqual(True, acme_ui.visible)

        return

    def test_load_and_save(self):
        """ load and save """

        p = self.preferences
        p.load(self.example)

        class AcmeUI(HasTraits):
            """ The Acme UI class! """

            # The traits that we want to initialize from preferences.
            bgcolor = Str('red')
            width   = Int(60)
            ratio   = Float(2.0)
            visible = Bool(False)

        acme_ui = AcmeUI()

        # Make some bindings.
        bind_preference(acme_ui, 'bgcolor', 'acme.ui.bgcolor')
        bind_preference(acme_ui, 'width',   'acme.ui.width')
        bind_preference(acme_ui, 'ratio',   'acme.ui.ratio')
        bind_preference(acme_ui, 'visible', 'acme.ui.visible')

        # Make sure the helper was initialized properly (with the values in
        # the loaded .ini file *not* the trait defaults!).
        self.assertEqual('blue', acme_ui.bgcolor)
        self.assertEqual(50, acme_ui.width)
        self.assertEqual(1.0, acme_ui.ratio)
        self.assertEqual(True, acme_ui.visible)

        # Make a change to one of the preference values.
        p.set('acme.ui.bgcolor', 'yellow')
        self.assertEqual('yellow', acme_ui.bgcolor)
        self.assertEqual('yellow', p.get('acme.ui.bgcolor'))

        # Save the preferences to a different file.
        tmpdir = tempfile.mkdtemp()
        tmp = join(tmpdir, 'tmp.ini')
        p.save(tmp)

        # Load the preferences again from that file.
        p = set_default_preferences(Preferences())
        p.load(tmp)

        acme_ui = AcmeUI()

        # Make some bindings.
        bind_preference(acme_ui, 'bgcolor', 'acme.ui.bgcolor')
        bind_preference(acme_ui, 'width',   'acme.ui.width')
        bind_preference(acme_ui, 'ratio',   'acme.ui.ratio')
        bind_preference(acme_ui, 'visible', 'acme.ui.visible')

        # Make sure the helper was initialized properly (with the values in
        # the .ini file *not* the trait defaults!).
        self.assertEqual('yellow', acme_ui.bgcolor)
        self.assertEqual(50, acme_ui.width)
        self.assertEqual(1.0, acme_ui.ratio)
        self.assertEqual(True, acme_ui.visible)

        # Clean up!
        os.remove(tmp)
        os.removedirs(tmpdir)

        return

    def test_explicit_preferences(self):
        """ explicit preferences """

        p = self.preferences
        p.load(self.example)

        class AcmeUI(HasTraits):
            """ The Acme UI class! """

            # The traits that we want to initialize from preferences.
            bgcolor = Str
            width   = Int
            ratio   = Float
            visible = Bool

        acme_ui = AcmeUI()
        acme_ui.on_trait_change(listener)

        # Create an empty preferences node and use that in some of the
        # bindings!
        preferences = Preferences()

        # Make some bindings.
        bind_preference(acme_ui, 'bgcolor', 'acme.ui.bgcolor', preferences)
        bind_preference(acme_ui, 'width',   'acme.ui.width')
        bind_preference(acme_ui, 'ratio',   'acme.ui.ratio', preferences)
        bind_preference(acme_ui, 'visible', 'acme.ui.visible')

        # Make sure the object was initialized properly.
        self.assertEqual('', acme_ui.bgcolor)
        self.assertEqual(50, acme_ui.width)
        self.assertEqual(0.0, acme_ui.ratio)
        self.assertEqual(True, acme_ui.visible)

        return

    def test_nested_set_in_trait_change_handler(self):
        """ nested set in trait change handler """

        p = self.preferences
        p.load(self.example)

        class AcmeUI(HasTraits):
            """ The Acme UI class! """

            # The traits that we want to initialize from preferences.
            bgcolor = Str
            width   = Int
            ratio   = Float
            visible = Bool

            def _width_changed(self, trait_name, old, new):
                """ Static trait change handler. """

                self.ratio = 3.0

                return

        acme_ui = AcmeUI()
        acme_ui.on_trait_change(listener)

        # Make some bindings.
        bind_preference(acme_ui, 'bgcolor', 'acme.ui.bgcolor')
        bind_preference(acme_ui, 'width',   'acme.ui.width')
        bind_preference(acme_ui, 'ratio',   'acme.ui.ratio')
        bind_preference(acme_ui, 'visible', 'acme.ui.visible')

        # Make sure the object was initialized properly.
        self.assertEqual('blue', acme_ui.bgcolor)
        self.assertEqual(50, acme_ui.width)
        self.assertEqual(1.0, acme_ui.ratio)
        self.assertEqual(True, acme_ui.visible)

        # Change the width via the preferences node. This should cause the
        # ratio to get set via the static trait change handler on the helper.
        p.set('acme.ui.width', 42)
        self.assertEqual(42, acme_ui.width)
        self.assertEqual('42', p.get('acme.ui.width'))

        # Did the ratio get changed?
        self.assertEqual(3.0, acme_ui.ratio)
        self.assertEqual('3.0', p.get('acme.ui.ratio'))

        return

    def test_trait_name_different_to_preference_name(self):

        p = self.preferences
        p.load(self.example)

        class AcmeUI(HasTraits):
            """ The Acme UI class! """

            # The test here is to have a different name for the trait than the
            # preference value (which is 'bgcolor').
            color = Str

        acme_ui = AcmeUI()
        acme_ui.on_trait_change(listener)

        # Make some bindings.
        bind_preference(acme_ui, 'color', 'acme.ui.bgcolor')

        # Make sure the object was initialized properly.
        self.assertEqual('blue', acme_ui.color)

        # Change the width via the preferences node.
        p.set('acme.ui.bgcolor', 'red')
        self.assertEqual('color', listener.trait_name)
        self.assertEqual('blue', listener.old)
        self.assertEqual('red', listener.new)

        return

        

# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2008 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#          Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
#-----------------------------------------------------------------------------

# Test cases.
import random
import pickle
import apptools.sweet_pickle as sweet_pickle

########################################

# Usecase1: generic case
class A(object):
    def __init__(self, b=None):
        self.x = 0
        self.set_b(b)

    def set_b(self, b):
        self.b_ref = b
        if b and hasattr(b, 'y'):
            self.x = b.y

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.set_b(self.b_ref)

    def __initialize__(self):
        while self.b_ref is None and self.b_ref.y != 0:
            yield True
        self.set_b(self.b_ref)


class B(object):
    def __init__(self, a=None):
        self.y = 0
        self.set_a(a)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['y']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.set_a(self.a_ref)

    def set_a(self, a):
        self.a_ref = a
        if a and hasattr(a, 'x'):
            self.y = a.x

    def __initialize__(self):
        while self.a_ref is None and self.a_ref.x != 0:
            yield True
        self.set_a(self.a_ref)


def test_generic():
    print('\nRunning generic test...')

    a = A()
    b = B()
    a.x = random.randint(1, 100)
    b.set_a(a)
    a.set_b(b)
    value = a.x

    # This will fail, even though we have a __setstate__ method.
    s = pickle.dumps(a)
    new_a = pickle.loads(s)
    try:
        print('\ta.x: %s' % new_a.x)
        print('\ta.b_ref.y: %s' % new_a.b_ref.y)
    except Exception as msg:
        print('\t%s' % 'Expected Error'.center(75,'*'))
        print('\t%s' % msg)
        print('\t%s' % ('*'*75))

    # This will work!
    s = pickle.dumps(a)
    new_a = sweet_pickle.loads(s)
    assert new_a.x == new_a.b_ref.y == value

    print('Generic test succesfull.\n\n')


########################################
# Usecase2: Toy Application
import re
class StringFinder(object):
    def __init__(self, source, pattern):
        self.pattern = pattern
        self.source = source
        self.data = []

    def __getstate__(self):
        s = self.__dict__.copy()
        del s['data']
        return s

    def __initialize__(self):
        while not self.source.initialized:
            yield True
        self.find()

    def find(self):
        pattern = self.pattern
        string = self.source.data
        self.data = [(x.start(), x.end()) for x in re.finditer(pattern, string)]


class XMLFileReader(object):
    def __init__(self, file_name):
        self.data = ''
        self.initialized = False
        self.file_name = file_name
        self.read()

    def __getstate__(self):
        s = self.__dict__.copy()
        del s['data']
        del s['initialized']
        return s

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.read()

    def read(self):
        # Make up random data from the filename
        data = [10*x for x in self.file_name]
        random.shuffle(data)
        self.data = ' '.join(data)
        self.initialized = True


class Application(object):
    def __init__(self):
        self.reader = XMLFileReader('some_test_file.xml')
        self.finder = StringFinder(self.reader, 'e')

    def get(self):
        print('\t%s' % self.finder.data)
        print('\t%s' % self.reader.data)


def test_toy_app():
    print('\nRunning toy app test...')

    a = Application()
    a.finder.find()
    a.get()
    s = pickle.dumps(a)
    b = pickle.loads(s)
    # Won't work.
    try:
        b.get()
    except Exception as msg:
        print('\t%s' % 'Expected Error'.center(75,'*'))
        print('\t%s' % msg)
        print('\t%s' % ('*'*75))

    # Works fine.
    c = sweet_pickle.loads(s)
    c.get()

    print('Toy app test succesfull.\n\n')


if __name__ == '__main__':
    test_generic()
    test_toy_app()
    print('ALL TESTS SUCCESFULL\n')



""" A binding between a trait on an object and a preference value. """


# Enthought library imports.
from traits.api import Any, HasTraits, Instance, Str, Undefined
from traits.api import Unicode

# Local imports.
from .i_preferences import IPreferences
from .package_globals import get_default_preferences


class PreferenceBinding(HasTraits):
    """ A binding between a trait on an object and a preference value. """

    #### 'PreferenceBinding' interface ########################################

    # The object that we are binding the preference to.
    obj = Any

    # The preferences node used by the binding. If this trait is not set then
    # the package-global default preferences node is used (and if that is not
    # set then the binding won't work ;^)
    preferences = Instance(IPreferences)

    # The path to the preference value.
    preference_path = Str

    # The name of the trait that we are binding the preference to.
    trait_name = Str

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """ Constructor. """

        super(PreferenceBinding, self).__init__(**traits)

        # Initialize the object's trait from the preference value.
        self._set_trait(notify=False)

        # Wire-up trait change handlers etc.
        self._initialize()

        return

    ###########################################################################
    # 'PreferenceBinding' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _preferences_default(self):
        """ Trait initializer. """

        return get_default_preferences()

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait change handlers ################################################

    def _on_trait_changed(self, obj, trait_name, old, new):
        """ Dynamic trait change handler. """

        self.preferences.set(self.preference_path, new)

        return

    #### Other observer pattern listeners #####################################

    def _preferences_listener(self, node, key, old, new):
        """ Listener called when a preference value is changed. """

        components = self.preference_path.split('.')
        if key == components[-1]:
            self._set_trait()

        return

    #### Methods ##############################################################

    # fixme: This method is mostly duplicated in 'PreferencesHelper' (the only
    # difference is the line that gets the handler).
    def _get_value(self, trait_name, value):
        """ Get the actual value to set.

        This method makes sure that any required work is done to convert the
        preference value from a string.

        """

        handler = self.obj.trait(trait_name).handler

        # If the trait type is 'Str' then we just take the raw value.
        if type(handler) is Str:
            pass

        # If the trait type is 'Unicode' then we convert the raw value.
        elif type(handler) is Unicode:
            value = str(value)

        # Otherwise, we eval it!
        else:
            try:
                value = eval(value)

            # If the eval fails then there is probably a syntax error, but
            # we will let the handler validation throw the exception.
            except:
                pass

        return handler.validate(self, trait_name, value)

    def _initialize(self):
        """ Wire-up trait change handlers etc. """

        # Listen for the object's trait being changed.
        self.obj.on_trait_change(self._on_trait_changed, self.trait_name)

        # Listen for the preference value being changed.
        components = self.preference_path.split('.')
        node       = '.'.join(components[:-1])

        self.preferences.add_preferences_listener(
            self._preferences_listener, node
        )

        return

    def _set_trait(self, notify=True):
        """ Set the object's trait to the value of the preference. """

        value = self.preferences.get(self.preference_path, Undefined)
        if value is not Undefined:
            trait_value = self._get_value(self.trait_name, value)
            traits      = {self.trait_name : trait_value}

            self.obj.set(trait_change_notify=notify, **traits)

        return


# Factory function for creating bindings.
def bind_preference(obj, trait_name, preference_path, preferences=None):
    """ Create a new preference binding. """

    # This may seem a bit wierd, but we manually build up a dictionary of
    # the traits that need to be set at the time the 'PreferenceBinding'
    # instance is created.
    #
    # This is because we only want to set the 'preferences' trait iff one
    # is explicitly specified. If we passed it in with the default argument
    # value of 'None' then it counts as 'setting' the trait which prevents
    # the binding instance from defaulting to the package-global preferences.
    # Also, if we try to set the 'preferences' trait *after* construction time
    # then it is too late as the binding initialization is done in the
    # constructor (we could of course split that out, which may be the 'right'
    # way to do it ;^).
    traits = {
        'obj'             : obj,
        'trait_name'      : trait_name,
        'preference_path' : preference_path
    }

    if preferences is not None:
        traits['preferences'] = preferences

    return PreferenceBinding(**traits)

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" Tests the default directory context. """


# Enthought library imports.
from apptools.naming.api import *

# Local imports.
from context_test_case import ContextTestCase


class DirContextTestCase(ContextTestCase):
    """ Tests the default directory context. """

    ###########################################################################
    # 'ContextTestCase' interface.
    ###########################################################################

    def create_context(self):
        """ Creates the context that we are testing. """

        return DirContext()

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_get_attributes(self):
        """ get attributes """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')
        self.assert_(isinstance(sub, DirContext))

        #### Generic name resolution tests ####

        # Non-existent name.
        self.assertRaises(NameNotFoundError, context.get_attributes, 'x')

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.get_attributes, 'x/a')

        # Attempt to resolve via an existing name that is not a context.
        context.bind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)
        self.assertRaises(NotContextError,context.get_attributes,'sub/a/x')

        #### Operation specific tests ####

        # Attributes of the root context.
        attributes = self.context.get_attributes('')
        self.assertEqual(len(attributes), 0)

        # Attributes of a sub-context.
        attributes = context.get_attributes('sub')
        self.assertEqual(len(attributes), 0)

        return

    def test_set_get_attributes(self):
        """ get and set attributes """

        defaults = {'colour' : 'blue'}

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')
        self.assert_(isinstance(sub, DirContext))

        #### Generic name resolution tests ####

        # Non-existent name.
        self.assertRaises(
            NameNotFoundError, context.set_attributes, 'x', defaults
        )

        # Attempt to resolve via a non-existent context.
        self.assertRaises(
            NameNotFoundError, context.set_attributes, 'x/a', defaults
        )

        # Attempt to resolve via an existing name that is not a context.
        context.bind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)
        self.assertRaises(
            NotContextError, context.set_attributes, 'sub/a/xx', defaults
        )

        #### Operation specific tests ####

        # Attributes of the root context.
        attributes = self.context.get_attributes('')
        self.assertEqual(len(attributes), 0)

        # Set the attributes.
        context.set_attributes('', defaults)
        attributes = context.get_attributes('')
        self.assertEqual(len(attributes), 1)
        self.assertEqual(attributes['colour'], 'blue')

        # Attributes of a sub-context.
        attributes = context.get_attributes('sub')
        self.assertEqual(len(attributes), 0)

        # Set the attributes.
        context.set_attributes('sub', defaults)
        attributes = context.get_attributes('sub')
        self.assertEqual(len(attributes), 1)
        self.assertEqual(attributes['colour'], 'blue')

        return

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The base class for all object serializers. """


# Standard library imports.
import logging
from traceback import print_exc
from os.path import splitext
#import cPickle
#import pickle

# Enthought library imports.
import apptools.sweet_pickle as sweet_pickle
from traits.api import HasTraits, Str


# Setup a logger for this module.
logger = logging.getLogger(__name__)


class ObjectSerializer(HasTraits):
    """ The base class for all object serializers. """

    #### 'ObjectSerializer' interface #########################################

    # The file extension recognized by this serializer.
    ext = Str('.pickle')

    ###########################################################################
    # 'ObjectSerializer' interface.
    ###########################################################################

    def can_load(self, path):
        """ Returns True if the serializer can load a file. """

        rest, ext = splitext(path)

        return ext == self.ext

    def load(self, path):
        """ Loads an object from a file. """

        # Unpickle the object.
        f = open(path, 'rb')
        try:
            try:
                 obj = sweet_pickle.load(f)
#                obj = cPickle.load(f)
#                obj = pickle.load(f)
            except Exception as ex:
                print_exc()
                logger.exception( "Failed to load pickle file: %s, %s" % (path, ex))

                raise
        finally:
            f.close()

        return obj

    def can_save(self, obj):
        """ Returns True if the serializer can save an object. """

        return True

    def save(self, path, obj):
        """ Saves an object to a file. """

        if not path.endswith(self.ext):
            actual_path = path + self.ext

        else:
            actual_path = path

        # Pickle the object.
        f = open(actual_path, 'wb')
        try:
            sweet_pickle.dump(obj, f, 1)
#            cPickle.dump(obj, f, 1)
#            pickle.dump(obj, f, 1)
        except Exception as ex:
            logger.exception( "Failed to pickle into file: %s, %s, object:%s"
                % (path, ex, obj))
            print_exc()
        f.close()

        return actual_path

### EOF #######################################################################



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2015, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# -----------------------------------------------------------------------------

from threading import RLock

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


from traits.api import Callable, Event, HasStrictTraits, Instance, Int


class LRUCache(HasStrictTraits):
    """ A least-recently used cache.

    Items older than `size()` accesses are dropped from the cache.

    """

    size = Int

    # Called with the key and value that was dropped from the cache
    cache_drop_callback = Callable

    # This event contains the set of cached cell keys whenever it changes
    updated = Event()

    _lock = Instance(RLock, args=())

    _cache = Instance(OrderedDict)

    def __init__(self, size, **traits):
        self.size = size
        self._initialize_cache()
        super(LRUCache, self).__init__(**traits)

    def _initialize_cache(self):
        with self._lock:
            if self._cache is None:
                self._cache = OrderedDict()
            else:
                self._cache.clear()

    def _renew(self, key):
        with self._lock:
            r = self._cache.pop(key)
            self._cache[key] = r
        return r

    # -------------------------------------------------------------------------
    # LRUCache interface
    # -------------------------------------------------------------------------

    def __contains__(self, key):
        with self._lock:
            return key in self._cache

    def __len__(self):
        with self._lock:
            return len(self._cache)

    def __getitem__(self, key):
        with self._lock:
            return self._renew(key)

    def __setitem__(self, key, result):
        try:
            dropped = None
            with self._lock:
                self._cache[key] = result
                self._renew(key)
                if self.size < len(self._cache):
                    dropped = self._cache.popitem(last=False)
            if dropped and self.cache_drop_callback is not None:
                self.cache_drop_callback(*dropped)
        finally:
            self.updated = list(self.keys())

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def items(self):
        with self._lock:
            return list(self._cache.items())

    def keys(self):
        with self._lock:
            return list(self._cache.keys())

    def values(self):
        with self._lock:
            return list(self._cache.values())

    def clear(self):
        with self._lock:
            self._initialize_cache()
        self.updated = []



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" Tests operations that span contexts. """


# Standard library imports.
import unittest

# Enthought library imports.
from apptools.naming.api import *


class ContextTestCase(unittest.TestCase):
    """ Tests naming operations that span contexts. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        self.context = self.create_context()
        self.context.create_subcontext('sub')

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        self.context = None

        return

    ###########################################################################
    # 'ContextTestCase' interface.
    ###########################################################################

    def create_context(self):
        """ Creates the context that we are testing. """

        return Context()

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_bind(self):
        """ bind """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')
        self.assert_(isinstance(sub, Context))

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.bind, '', 1)

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.bind, 'xx/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Bind a name.
        context.bind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Attempt to resolve via an existing name that is not a context.
        self.assertRaises(NotContextError, context.bind, 'sub/a/xx', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Try to bind it again.
        self.assertRaises(NameAlreadyBoundError, context.bind, 'sub/a', 1)

        return

    def test_bind_with_make_contexts(self):
        """ bind with make contexts """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')
        self.assert_(isinstance(sub, Context))

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.bind, '', 1, True)

        # Attempt to resolve via a non-existent context - which should result
        # in the context being created automatically.
        context.bind('xx/a', 1, True)
        self.assertEqual(len(context.list_bindings('xx')), 1)
        self.assertEqual(1, context.lookup('xx/a'))

        # Bind an even more 'nested' name.
        context.bind('xx/foo/bar/baz', 42, True)
        self.assertEqual(len(context.list_bindings('xx/foo/bar')), 1)
        self.assertEqual(42, context.lookup('xx/foo/bar/baz'))

        return

    def test_rebind(self):
        """ context rebind """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.rebind, '', 1)

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.rebind, 'xx/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Bind a name.
        context.bind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Rebind it.
        context.rebind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Attempt to resolve via an existing name that is not a context.
        self.assertRaises(NotContextError, context.rebind, 'sub/a/xx', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        return

    def test_rebind_with_make_contexts(self):
        """ rebind with make contexts """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')
        self.assert_(isinstance(sub, Context))

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.rebind, '', 1, True)

        # Attempt to resolve via a non-existent context - which should result
        # in the context being created automatically.
        context.rebind('xx/a', 1, True)
        self.assertEqual(len(context.list_bindings('xx')), 1)
        self.assertEqual(1, context.lookup('xx/a'))

        # Rebind an even more 'nested' name.
        context.rebind('xx/foo/bar/baz', 42, True)
        self.assertEqual(len(context.list_bindings('xx/foo/bar')), 1)
        self.assertEqual(42, context.lookup('xx/foo/bar/baz'))

        # And do it again... (this is REbind after all).
        context.rebind('xx/foo/bar/baz', 42, True)
        self.assertEqual(len(context.list_bindings('xx/foo/bar')), 1)
        self.assertEqual(42, context.lookup('xx/foo/bar/baz'))

        return

    def test_unbind(self):
        """ context unbind """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.unbind, '')

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.unbind, 'xx/a')
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Bind a name.
        context.bind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Attempt to resolve via an existing name that is not a context.
        self.assertRaises(NotContextError, context.unbind, 'sub/a/xx')
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Unbind it.
        context.unbind('sub/a')
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Try to unbind a non-existent name.
        self.assertRaises(NameNotFoundError, context.unbind, 'sub/b')

        return

    def test_rename_object(self):
        """ rename an object """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.rename, '', 'x')
        self.assertRaises(InvalidNameError, context.rename, 'x', '')

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.rename, 'x/a', 'x/b')
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Bind a name.
        context.bind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Attempt to resolve via an existing name that is not a context.
        self.assertRaises(NotContextError, context.bind, 'sub/a/xx', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Rename it.
        context.rename('sub/a', 'sub/b')
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Lookup using the new name.
        self.assertEqual(context.lookup('sub/b'), 1)

        # Lookup using the old name.
        self.assertRaises(NameNotFoundError, context.lookup, 'sub/a')

        return

    def test_rename_context(self):
        """ rename a context """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.rename, '', 'x')
        self.assertRaises(InvalidNameError, context.rename, 'x', '')

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.rename, 'x/a', 'x/b')
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Create a context.
        context.create_subcontext('sub/a')
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Rename it.
        context.rename('sub/a', 'sub/b')
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Lookup using the new name.
        context.lookup('sub/b')

        # Lookup using the old name.
        self.assertRaises(NameNotFoundError, context.lookup, 'sub/a')

        return

    def test_lookup(self):
        """ lookup """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.lookup, 'xx/a')
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Bind a name.
        context.bind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Attempt to resolve via an existing name that is not a context.
        self.assertRaises(NotContextError, context.lookup, 'sub/a/xx')
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Look it up.
        self.assertEqual(context.lookup('sub/a'), 1)

        # Looking up the Empty name returns the context itself.
        self.assertEqual(context.lookup(''), context)

        # Non-existent name.
        self.assertRaises(NameNotFoundError, context.lookup, 'sub/b')

        return

    def test_create_subcontext(self):
        """ create sub-context """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.create_subcontext, '')

        # Attempt to resolve via a non-existent context.
        self.assertRaises(
            NameNotFoundError, context.create_subcontext, 'xx/a'
        )
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Create a sub-context.
        context.create_subcontext('sub/a')
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Try to bind it again.
        self.assertRaises(
            NameAlreadyBoundError, context.create_subcontext, 'sub/a'
        )

        # Bind a name.
        context.bind('sub/b', 1)
        self.assertEqual(len(sub.list_bindings('')), 2)

        # Attempt to resolve via an existing name that is not a context.
        self.assertRaises(
            NotContextError, context.create_subcontext, 'sub/b/xx'
        )
        self.assertEqual(len(sub.list_bindings('')), 2)

        return

    def test_destroy_subcontext(self):
        """ single context destroy sub-context """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.destroy_subcontext, '')

        # Attempt to resolve via a non-existent context.
        self.assertRaises(
            NameNotFoundError, context.destroy_subcontext, 'xx/a'
        )
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Create a sub-context.
        context.create_subcontext('sub/a')
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Destroy it.
        context.destroy_subcontext('sub/a')
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Bind a name.
        context.bind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Try to destroy it.
        self.assertRaises(
            NotContextError, context.destroy_subcontext, 'sub/a'
        )

        # Try to destroy a non-existent name.
        self.assertRaises(
            NameNotFoundError, context.destroy_subcontext, 'sub/b'
        )

        # Attempt to resolve via an existing name that is not a context.
        self.assertRaises(
            NotContextError, context.destroy_subcontext, 'sub/a/xx'
        )
        self.assertEqual(len(sub.list_bindings('')), 1)

        return

    def test_list_bindings(self):
        """ list bindings """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')

        # List the bindings in the root.
        bindings = context.list_bindings('')
        self.assertEqual(len(bindings), 1)

        # List the names in the sub-context.
        bindings = context.list_bindings('sub')
        self.assertEqual(len(bindings), 0)

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.list_bindings, 'xx/a')
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Bind a name.
        context.bind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Attempt to resolve via an existing name that is not a context.
        self.assertRaises(
            NotContextError, context.list_bindings, 'sub/a/xx'
        )
        self.assertEqual(len(sub.list_bindings('')), 1)

        return

    def test_list_names(self):
        """ list names """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')

        # List the names in the root.
        names = context.list_names('')
        self.assertEqual(len(names), 1)

        # List the names in the sub-context.
        names = context.list_names('sub')
        self.assertEqual(len(names), 0)

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.list_names, 'xx/a')
        self.assertEqual(len(sub.list_bindings('')), 0)

        # Bind a name.
        context.bind('sub/a', 1)
        self.assertEqual(len(sub.list_bindings('')), 1)

        # Attempt to resolve via an existing name that is not a context.
        self.assertRaises(
            NotContextError, context.list_names, 'sub/a/xx'
        )
        self.assertEqual(len(sub.list_bindings('')), 1)

        return

    def test_default_factories(self):
        """ default object and state factories. """

        object_factory = ObjectFactory()
        self.assertRaises(
            NotImplementedError, object_factory.get_object_instance, 0, 0, 0
        )

        state_factory = StateFactory()
        self.assertRaises(
            NotImplementedError, state_factory.get_state_to_bind, 0, 0, 0
        )

        return

    def test_search(self):
        """ test retrieving the names of bound objects """

        # Convenience.
        context = self.context
        sub = self.context.lookup('sub')
        sub_sibling = context.create_subcontext('sub sibling')
        sub_sub = sub.create_subcontext('sub sub')

        context.bind('one',1)
        names = context.search( 1 )
        self.assertEqual( len(names), 1)
        self.assertEqual( names[0], 'one' )

        names = sub.search(1)
        self.assertEqual( len(names), 0)

        context.bind('sub/two',2)
        names = context.search( 2 )
        self.assertEqual( len(names), 1)
        self.assertEqual( names[0], 'sub/two' )

        names = sub.search( 2 )
        self.assertEqual( len(names), 1)
        self.assertEqual( names[0], 'two' )

        context.bind('sub/sub sub/one',1)
        names = context.search( 1 )
        self.assertEqual( len(names), 2)
        self.assertEqual( sorted(names), sorted(['one', 'sub/sub sub/one']) )

        names = sub.search(None)
        self.assertEqual( len(names), 0)

        names = context.search( sub_sub )
        self.assertEqual( len(names), 1)
        self.assertEqual( names[0], 'sub/sub sub' )

        return

#### EOF ######################################################################
